from __future__ import annotations
from functools import lru_cache
import numpy as np
from acryo.backend import Backend, AnyArray


@lru_cache(maxsize=10)
def _build_mesh(
    shape: tuple[int, int, int],
    max_shifts: tuple[float, float, float],
    upsample: int,
    backend: Backend,
) -> AnyArray[np.float32]:
    upsampled_max_shifts = (np.asarray(max_shifts) * upsample).astype(np.int32)
    center = np.array(shape) / 2 - 0.5
    mesh = backend.meshgrid(
        *[
            backend.linspace(c - width / upsample, c + width / upsample, 2 * width + 1)
            for c, width in zip(center, upsampled_max_shifts)
        ],
        indexing="ij",
    )
    return backend.stack(mesh, axis=0)


def build_mesh(
    shape: tuple[int, ...],
    max_shifts: tuple[float, ...],
    upsample: int,
    backend: Backend,
) -> AnyArray[np.float32]:
    """Build a meshgrid for up-sampling.

    Parameters
    ----------
    shape : tuple[int, int, int]
        Shape of the original volume from which up-sampling will be applied.
    max_shifts : tuple[float, float, float]
        Maximum shifts in each direction in pixel.
    upsample : int
        Up-sampling factor.
    """
    return _build_mesh(shape, max_shifts, upsample, backend)



# pyright: reportPrivateImportUsage=false
from __future__ import annotations

from typing import NamedTuple, Mapping
import tempfile
from contextlib import contextmanager
import numpy as np
from dask import array as da
from acryo.backend import Backend


class _Subtomogram(NamedTuple):
    array: da.Array
    backend: str


class SubtomogramCache(Mapping[int, _Subtomogram]):
    def __init__(self, dir: str | None = None) -> None:
        self._dict: dict[int, _Subtomogram] = {}
        self._cache_dir = dir

    def __getitem__(self, id_: int) -> _Subtomogram:
        return self._dict[id_]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def cache_array(self, dsk: da.Array, id_: int) -> da.Array:
        shape = dsk.shape
        with tempfile.NamedTemporaryFile(dir=self._cache_dir) as ntf:
            mmap = np.memmap(ntf, dtype=np.float32, mode="w+", shape=shape)

        component = dsk[(0,) * dsk.ndim].compute()
        _is_cupy = type(component).__module__.startswith("cupy")
        if _is_cupy:
            dsk = dsk.map_blocks(lambda x: x.get(), dtype=dsk.dtype)  # type: ignore
        da.store(dsk, mmap, compute=True)

        darr = da.from_array(
            mmap,
            chunks=("auto",) + shape[1:],  # type: ignore
            meta=np.array([], dtype=np.float32),
        )
        self._dict[id_] = _Subtomogram(darr, "cupy" if _is_cupy else "numpy")
        return darr

    def get_cache(
        self,
        id_: int,
        shape: tuple[int, int, int] | None,
        backend: Backend | None = None,
    ) -> da.Array | None:
        if id_ in self._dict:
            cache = self._dict[id_]
            dsk = cache.array
            if shape is None or dsk.shape[1:] == shape:
                if backend is None:
                    return dsk
                elif cache.backend == "cupy" and backend.name == "numpy":
                    return dsk.map_blocks(backend.asnumpy, dtype=dsk.dtype)  # type: ignore
                return dsk.map_blocks(backend.asarray, dtype=dsk.dtype)  # type: ignore
        return None

    def delete_cache(self, id_):
        self._dict.pop(id_, None)

    @contextmanager
    def temporal(self):
        old_state = self._dict.copy()
        try:
            yield
        finally:
            self._dict = old_state



from .core import single_axis, dual_axis, no_wedge
from ._base import TiltSeriesModel, NoWedge, UnionAxes
from ._single import SingleAxis

__all__ = [
    "single_axis",
    "dual_axis",
    "no_wedge",
    "TiltSeriesModel",
    "SingleAxis",
    "NoWedge",
    "UnionAxes",
]



from __future__ import annotations
import itertools
from typing import Callable, Literal, Sequence, cast
import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation

from acryo._types import Ranges, RangeLike, RotationType
from acryo.molecules import from_euler_xyz_coords
from acryo._typed_scipy import affine_transform


def _normalize_a_range(rng: RangeLike) -> RangeLike:
    if len(rng) != 2:
        raise TypeError(f"Range must be defined by (float, float), got {rng!r}")
    max_rot, drot = rng
    return float(max_rot), float(drot)


def _normalize_ranges(rng: RangeLike | Ranges) -> Ranges:
    if np.array(rng).ndim == 2:
        return tuple(_normalize_a_range(r) for r in rng)  # type: ignore
    else:
        rng_ = _normalize_a_range(rng)  # type: ignore
        return (rng_,) * 3


def _seq_of_max_and_step_to_quat(rotations: RangeLike | Ranges) -> NDArray[np.float32]:
    _rotations = _normalize_ranges(rotations)
    angles = []
    for max_rot, step in _rotations:
        if step == 0:
            angles.append(np.zeros(1))
        else:
            n = int(max_rot / step)
            angles.append(np.linspace(-n * step, n * step, 2 * n + 1))

    _quat: list[NDArray[np.float32]] = []
    for angs in itertools.product(*angles):
        _quat.append(
            from_euler_xyz_coords(np.array(angs), "zyx", degrees=True)
            .as_quat(canonical=False)
            .astype(np.float32, copy=False)
        )
    return np.stack(_quat, axis=0)


def normalize_rotations(rotations: RotationType | None) -> NDArray[np.floating]:
    """
    Normalize various rotation expressions to quaternions.

    Parameters
    ----------
    rotations : tuple of float and float, or list of it, optional
        Rotation around each axis.

    Returns
    -------
    np.ndarray
        Corresponding quaternions in shape (N, 4).
    """
    if isinstance(rotations, Rotation):
        quats = rotations.as_quat(canonical=False)
    elif rotations is not None:
        if not hasattr(rotations, "__iter__"):
            raise TypeError("rotations must be iterable")
        list_rot = list(rotations)
        if isinstance(list_rot[0], Rotation):
            list_rot = cast("list[Rotation]", list_rot)
            quats = np.stack(
                [
                    r.as_quat(canonical=False).astype(np.float32, copy=False)
                    for r in list_rot
                ],
                axis=0,
            )
        else:
            quats = _seq_of_max_and_step_to_quat(rotations)
    else:
        quats = np.array([[0.0, 0.0, 0.0, 1.0]], dtype=np.float32)

    return quats


def rotate(
    image: np.ndarray,
    degrees: tuple[float, float, float] | Sequence[float],
    order: int = 3,
    mode: Literal["constant", "nearest", "mirror", "wrap", "reflect"] = "constant",
    cval: Callable | float = np.mean,
):
    from acryo._utils import compose_matrices

    quat = euler_to_quat(degrees)
    rotator = Rotation.from_quat(quat).inv()
    matrix = compose_matrices(
        np.array(image.shape[-image.ndim :]) / 2 - 0.5, [rotator]
    )[0]
    if callable(cval):
        _cval = cval(image)
    else:
        _cval = cval

    return affine_transform(
        image,
        matrix=matrix,
        order=order,
        mode=mode,
        cval=_cval,
        prefilter=order > 1,
    )


def euler_to_quat(degrees):
    return from_euler_xyz_coords(np.array(degrees), "zyx", degrees=True).as_quat(
        canonical=False
    )



# pyright: reportPrivateImportUsage=false
from __future__ import annotations

from contextlib import contextmanager
from abc import ABC, abstractmethod, abstractproperty
from typing import (
    Callable,
    TYPE_CHECKING,
    Iterable,
    Iterator,
    NamedTuple,
    Any,
    Sequence,
    SupportsIndex,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import TypeGuard
import warnings
import numpy as np
from numpy.typing import NDArray
from dask import array as da
from dask.delayed import delayed
from scipy.spatial.transform import Rotation
import polars as pl

from acryo.alignment import (
    BaseAlignmentModel,
    ZNCCAlignment,
    RotationImplemented,
    AlignmentFactory,
    AlignmentResult,
)
from acryo import _utils
from acryo._types import nm, pixel
from acryo.backend import Backend
from acryo._dask import DaskArrayList, DaskTaskList, DaskTaskIterator, compute
from acryo.molecules import Molecules
from acryo.loader import _misc
from acryo.loader._group import LoaderGroup
from acryo.loader._cache import SubtomogramCache
from acryo.pipe._classes import ImageProvider, ImageConverter

if TYPE_CHECKING:
    from typing_extensions import Self
    from acryo.classification import PcaClassifier
    from acryo.alignment._base import MaskType

_R = TypeVar("_R")

_ShapeType = Union[pixel, tuple[pixel, ...], None]
TemplateInputType = Union[NDArray[np.float32], ImageProvider[np.ndarray]]
MaskInputType = Union[
    NDArray[np.float32], ImageProvider[np.ndarray], ImageConverter, None
]
AggFunction = Callable[[NDArray[np.float32]], _R]
IntoExpr = Union[str, pl.Expr]


class Unset:
    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "Unset"


CACHE = SubtomogramCache()


class LoaderBase(ABC):
    """The base class for subtomogram loaders."""

    def __init__(
        self,
        order: int = 3,
        scale: nm = 1.0,
        output_shape: pixel | tuple[pixel, pixel, pixel] | Unset = Unset(),
        corner_safe: bool = False,
    ) -> None:
        ndim = 3
        self._order, self._output_shape, self._scale, self._corner_safe = check_input(
            order, output_shape, scale, corner_safe, ndim
        )

    _CACHE = CACHE

    @abstractproperty
    def molecules(self) -> Molecules:
        """All the molecules"""
        raise NotImplementedError

    @property
    def features(self) -> pl.DataFrame:
        """Collect all the features from all the molecules"""
        return self.molecules.features

    @property
    def scale(self) -> nm:
        """Get the physical scale of tomogram."""
        return self._scale

    @property
    def output_shape(self) -> tuple[pixel, pixel, pixel] | Unset:
        """Return the output subtomogram shape."""
        return self._output_shape

    @property
    def order(self) -> int:
        """Return the interpolation order."""
        return self._order

    @property
    def corner_safe(self) -> bool:
        """Whether rotation is corner-safe."""
        return self._corner_safe

    @abstractmethod
    def construct_loading_tasks(
        self,
        output_shape: _ShapeType = None,
        backend: Backend | None = None,
    ) -> DaskArrayList:
        ...

    def _get_cached_array(
        self, shape: tuple[int, int, int] | None, backend: Backend | None
    ) -> da.Array | None:
        return CACHE.get_cache(id(self), shape, backend)

    def construct_dask(
        self,
        output_shape: pixel | tuple[pixel, ...] | None = None,
        backend: Backend | None = None,
    ) -> da.Array:
        """
        Construct a dask array of subtomograms.

        This function is always needed before parallel processing. If subtomograms
        are cached in a memory-map it will be used instead.

        Returns
        -------
        da.Array
            An 4-D array which ``arr[i]`` corresponds to the ``i``-th subtomogram.
        """
        output_shape = self._get_output_shape(output_shape)
        xp = backend or Backend()
        if (cached := self._get_cached_array(output_shape, xp)) is not None:
            return cached
        tasks = self.construct_loading_tasks(output_shape, xp)
        out = da.stack(tasks, axis=0)
        return out

    @abstractmethod
    def replace(
        self,
        molecules: Molecules | None = None,
        output_shape: pixel | tuple[pixel, pixel, pixel] | Unset | None = None,
        order: int | None = None,
        scale: float | None = None,
        corner_safe: bool | None = None,
    ) -> Self:
        ...

    def copy(self) -> Self:
        """Return a copy of the loader."""
        return self.replace()

    def count(self) -> int:
        """Return the number of subtomograms."""
        return len(self.molecules)

    def iter_mapping_tasks(
        self,
        func: Callable[..., _R],
        *const_args,
        output_shape: _ShapeType = None,
        var_kwarg: dict[str, Iterable[Any]] | None = None,
        **const_kwargs,
    ) -> DaskTaskIterator[_R]:
        """
        Iterate over delayed mapping tasks using subtomograms.

        Parameters
        ----------
        func : Callable
            Mapping function. The first argument of the function is the subtomogram.
        output_shape : int or tuple of int, optional
            Shape of subtomograms.
        var_kwarg : dict, optional
            Variable keyword arguments. The length of each argument must be the same
            as the number of subtomograms.

        Yields
        ------
        da.delayed.Delayed object
            Delayed tasks that are ready for ``da.compute``.
        """
        output_shape = self._get_output_shape(output_shape)
        dask_array = self.construct_loading_tasks(output_shape=output_shape)
        delayed_f = delayed(func)
        if var_kwarg is None:
            it = (delayed_f(ar, *const_args, **const_kwargs) for ar in dask_array)
        else:
            it = (
                delayed_f(ar, *const_args, **const_kwargs, **kw)
                for ar, kw in zip(dask_array, _misc.dict_iterrows(var_kwarg))
            )
        return DaskTaskIterator(it)

    def construct_mapping_tasks(
        self,
        func: Callable[..., _R],
        *const_args,
        output_shape: _ShapeType = None,
        var_kwarg: dict[str, Iterable[Any]] | None = None,
        **const_kwargs,
    ) -> DaskTaskList[_R]:
        """
        Construct delayed mapping tasks using subtomograms.

        Parameters
        ----------
        func : Callable
            Mapping function. The first argument of the function is the subtomogram.
        output_shape : int or tuple of int, optional
            Shape of subtomograms.
        var_kwarg : dict, optional
            Variable keyword arguments. The length of each argument must be the same
            as the number of subtomograms.

        Returns
        -------
        DaskTaskList
            List of delayed tasks that are ready for computation.
        """
        return self.iter_mapping_tasks(
            func,
            *const_args,
            output_shape=output_shape,
            var_kwarg=var_kwarg,
            **const_kwargs,
        ).tolist()

    def _create_cache(self, output_shape: _ShapeType = None) -> da.Array:
        """
        Create cached stack of subtomograms.

        Parameters
        ----------
        path : str, optional
            File path of the temporary file. If not given file will be created by
            ``tempfile.NamedTemporaryFile`` function.

        Returns
        -------
        da.Array
            A lazy-loading array that uses the memory-mapped array.
        """
        output_shape = self._get_output_shape(output_shape)
        backend = Backend()
        if (cached := self._get_cached_array(output_shape, backend)) is not None:
            return cached
        dsk = self.construct_dask(output_shape=output_shape, backend=backend)
        return CACHE.cache_array(dsk, id(self))

    @contextmanager
    def cached(self, output_shape: _ShapeType = None) -> Iterator[da.Array]:
        """
        Context manager for caching subtomograms of give shape.

        Subtomograms are cached in a temporary memory-map file. Within this context
        loading subtomograms of given output shape will be faster.
        """
        with CACHE.temporal():
            yield self._create_cache(output_shape=output_shape)

    def asnumpy(self, output_shape: _ShapeType = None) -> NDArray[np.float32]:
        """Load all the subtomograms as a 4D numpy array."""
        return self.construct_dask(output_shape=output_shape).compute()

    def load(
        self,
        idx: SupportsIndex | slice | Iterable[SupportsIndex],
        output_shape: _ShapeType = None,
    ):
        """
        Load subtomogram(s) of given index.

        Parameters
        ----------
        idx : int, slice or iterable of int
            Subtomogram index.
        output_shape : int or tuple of int, optional
            Shape of the output subtomograms.

        Returns
        -------
        3D array or 4D array
            Subtomogram(s) of given index.
        """
        tasks = self.construct_loading_tasks(output_shape=output_shape)
        if isinstance(idx, SupportsIndex):
            return tasks[idx].compute()
        elif isinstance(idx, slice):
            return da.stack(tasks[idx], axis=0).compute()
        elif hasattr(idx, "__iter__"):
            return da.stack([tasks[i] for i in idx], axis=0).compute()
        else:
            raise TypeError(f"Invalid index type: {type(idx)}")

    def load_iter(
        self, output_shape: _ShapeType = None
    ) -> Iterator[NDArray[np.float32]]:
        """
        Iterate over subtomograms.

        Parameters
        ----------
        output_shape : int or tuple of int, optional
            Shape of the output subtomograms.

        Yields
        ------
        3D array
            Subtomogram
        """
        tasks = self.construct_loading_tasks(output_shape=output_shape)
        for task in tasks:
            yield task.compute()

    def average(
        self, output_shape: _ShapeType = None, *, backend: Backend | None = None
    ) -> NDArray[np.float32]:
        """
        Calculate the average of subtomograms.

        This function execute so-called "subtomogram averaging". The size of
        subtomograms is determined by the ``self.output_shape`` attribute.

        Returns
        -------
        np.ndarray
            Averaged image
        """
        xp = backend or Backend()
        output_shape = self._get_output_shape(output_shape)
        dsk = self.construct_dask(output_shape=output_shape, backend=xp)
        dsk = dsk.rechunk(("auto",) + output_shape)  # type: ignore
        return xp.asnumpy(dsk.mean(axis=0).compute())

    def average_split(
        self,
        n_set: int = 1,
        seed: int | None = 0,
        squeeze: bool = True,
        output_shape: _ShapeType = None,
    ) -> NDArray[np.float32]:
        """
        Split subtomograms into two set and average separately.

        This method executes pairwise subtomogram averaging using randomly
        selected molecules, which is useful for calculation of such as Fourier
        shell correlation.

        Parameters
        ----------
        n_set : int, default is 1
            Number of split set of averaged image.
        seed : random seed, default is 0
            Random seed to determine how subtomograms will be split.
        squeeze : bool, default is True
            If true and n_set is 1, return a 4D array.
        output_shape : tuple of int, optional
            Output shape of the averaged image. If not given, the default output
            shape of the loader object will be used.

        Returns
        -------
        np.ndarray
            Averaged images. The shape of the array is (n_set, 2, *output_shape).
        """
        backend = Backend()
        output_shape = self._get_output_shape(output_shape)
        rng = np.random.default_rng(seed=seed)

        tasks: list[da.Array] = []
        dsk = self.construct_dask(output_shape=output_shape)
        nmole = dsk.shape[0]
        for _ in range(n_set):
            ind0, ind1 = _misc.random_splitter(rng, nmole)
            _stack = da.stack(
                [
                    dsk[ind0].rechunk(("auto",) + output_shape).mean(axis=0),  # type: ignore
                    dsk[ind1].rechunk(("auto",) + output_shape).mean(axis=0),  # type: ignore
                ],
                axis=0,
            )
            tasks.append(_stack)

        out = da.compute(tasks)[0]
        stack = np.stack([backend.asnumpy(a) for a in out], axis=0)
        if squeeze and n_set == 1:
            stack: NDArray[np.float32] = stack[0]
        return stack

    def align(
        self,
        template: TemplateInputType,
        *,
        mask: MaskInputType = None,
        max_shifts: nm | tuple[nm, nm, nm] = 1.0,
        alignment_model: type[BaseAlignmentModel] | AlignmentFactory = ZNCCAlignment,
        backend: Backend | None = None,
        **align_kwargs,
    ) -> Self:
        """
        Align subtomograms to the template image.

        This method conduct so called "subtomogram alignment". Only shifts and rotations
        are calculated in this method. To get averaged image, you'll have to run "average"
        method using the resulting SubtomogramLoader instance.

        Parameters
        ----------
        template : 3D array or ImageProvider
            Template image. If ImageProvider is given, the image will be provided
            accordingly using the scale of the loader object.
        mask : np.ndarray or callable of np.ndarray to np.ndarray optional
            Mask image. Must in the same shape as the template.
        max_shifts : int or tuple of int, default is (1., 1., 1.)
            Maximum shift between subtomograms and template.
        alignment_model : subclass of BaseAlignmentModel, optional
            Alignment model class used for subtomogram alignment. By default,
            ``ZNCCAlignment`` will be used.
        align_kwargs : optional keyword arguments
            Additional keyword arguments passed to the input alignment model.

        Returns
        -------
        subtomogram loader object
            A loader instance of the same type with updated molecules.
        """
        _backend = backend or Backend()
        max_shifts = _normalize_max_shifts(max_shifts)

        model = alignment_model(
            self.normalize_template(template),
            self.normalize_mask(mask),
            **align_kwargs,
        )
        if model.has_hetero_templates:
            return self.align_multi_templates(
                list(model.template),
                mask=mask,
                max_shifts=max_shifts,
                alignment_model=alignment_model,
                backend=backend,
                **align_kwargs,
            )

        _max_shifts_px = tuple(np.asarray(max_shifts) / self.scale)
        tasks = self.construct_mapping_tasks(
            model.align,
            max_shifts=_max_shifts_px,
            output_shape=model.input_shape,
            backend=_backend,
            var_kwarg=dict(
                quaternion=self.molecules.quaternion(),
                pos=self.molecules.pos / self.scale,
            ),
        )
        all_results = tasks.compute()
        return self._post_align(all_results, model.input_shape)

    def _post_align(
        self,
        results: list[AlignmentResult],
        shape: tuple[int, int, int],
    ) -> Self:
        local_shifts, local_rot, scores = _misc.allocate(len(results))
        for i, result in enumerate(results):
            _, loc_shift, local_rot[i], scores[i] = result
            local_shifts[i] = loc_shift * self.scale

        rotator = Rotation.from_quat(local_rot)
        mole_aligned = self.molecules.linear_transform(local_shifts, rotator)

        mole_aligned.features = self.molecules.features.with_columns(
            _misc.get_feature_list(scores, local_shifts, rotator.as_rotvec()),
        )

        return self.replace(molecules=mole_aligned, output_shape=shape)

    def align_no_template(
        self,
        *,
        mask: MaskInputType = None,
        max_shifts: nm | tuple[nm, nm, nm] = 1.0,
        output_shape: _ShapeType = None,
        alignment_model: type[BaseAlignmentModel] = ZNCCAlignment,
        **align_kwargs,
    ) -> Self:
        """
        Align subtomograms without template image.

        A template-free version of :func:`align`. This method first
        calculates averaged image and uses it for the alignment template. To
        avoid loading same subtomograms twice, a memory-mapped array is created
        internally (so the second subtomogram loading is faster).

        Parameters
        ----------
        mask : np.ndarray or callable, optional
            Mask image. Must in the same shape as the template.
        max_shifts : int or tuple of int, default is (1., 1., 1.)
            Maximum shift between subtomograms and template.
        alignment_model : subclass of BaseAlignmentModel, optional
            Alignment model class used for subtomogram alignment. By default,
            ``ZNCCAlignment`` will be used.
        align_kwargs : optional keyword arguments
            Additional keyword arguments passed to the input alignment model.

        Returns
        -------
        subtomogram loader object
            A loader instance of the same type with updated molecules.
        """
        if output_shape is None and isinstance(mask, np.ndarray):
            output_shape = mask.shape
        backend = Backend()
        with self.cached(output_shape=output_shape):
            template = self.average(output_shape=output_shape, backend=backend)
            out = self.align(
                template,
                mask=mask,
                max_shifts=max_shifts,
                alignment_model=alignment_model,
                backend=backend,
                **align_kwargs,
            )
        return out

    def align_multi_templates(
        self,
        templates: list[TemplateInputType] | ImageProvider[list[NDArray[np.float32]]],
        *,
        mask: MaskInputType = None,
        max_shifts: nm | tuple[nm, nm, nm] = 1.0,
        alignment_model: type[BaseAlignmentModel] | AlignmentFactory = ZNCCAlignment,
        backend: Backend | None = None,
        label_name: str = "labels",
        **align_kwargs,
    ) -> Self:
        """
        Align subtomograms with multiple template images.

        A multi-template version of :func:`align`. This method calculate cross
        correlation for every template and uses the best local shift, rotation and
        template.

        Parameters
        ----------
        templates: list of 3D arrays or ImageProvider
            Template images. If ImageProvider is given, the image will be provided
            accordingly using the scale of the loader object.
        mask : np.ndarray, optional
            Mask image. Must in the same shape as the template.
        max_shifts : int or tuple of int, default is (1., 1., 1.)
            Maximum shift between subtomograms and template.
        alignment_model : subclass of BaseAlignmentModel, optional
            Alignment model class used for subtomogram alignment. By default,
            ``ZNCCAlignment`` will be used.
        align_kwargs : optional keyword arguments
            Additional keyword arguments passed to the input alignment model.

        Returns
        -------
        subtomogram loader object
            A loader instance of the same type with updated molecules.
        """
        _backend = backend or Backend()
        _max_shifts_px = np.asarray(max_shifts) / self.scale

        if isinstance(templates, ImageProvider):
            _templates = templates(self.scale)
        else:
            _templates = [self.normalize_template(t) for t in templates]
        model = alignment_model(
            template=_templates,
            mask=self.normalize_mask(mask),
            **align_kwargs,
        )
        tasks = self.construct_mapping_tasks(
            model.align,
            max_shifts=_max_shifts_px,
            output_shape=model.input_shape,
            backend=_backend,
            var_kwarg=dict(
                quaternion=self.molecules.quaternion(),
                pos=self.molecules.pos / self.scale,
            ),
        )
        all_results = tasks.compute()
        if isinstance(model, RotationImplemented) and model._n_rotations > 1:
            remainder = len(_templates)
        else:
            remainder = -1
        return self._post_align_multi_templates(
            all_results, model.input_shape, remainder, label_name
        )

    def _post_align_multi_templates(
        self,
        results: list[AlignmentResult],
        shape: tuple[int, int, int],
        remainder: int = -1,
        label_name: str = "labels",
    ):
        local_shifts, local_rot, scores = _misc.allocate(len(results))
        labels = np.zeros(len(results), dtype=np.uint32)
        for i, result in enumerate(results):
            labels[i], loc_shift, local_rot[i], scores[i] = result
            local_shifts[i] = loc_shift * self.scale

        rotator = Rotation.from_quat(local_rot)
        mole_aligned = self.molecules.linear_transform(local_shifts, rotator)

        if remainder > 1:
            labels %= remainder  # type: ignore
        labels = labels.astype(np.uint8)

        feature_list = _misc.get_feature_list(scores, local_shifts, rotator.as_rotvec())
        mole_aligned.features = self.molecules.features.with_columns(
            feature_list + [pl.Series(label_name, labels)]
        )

        return self.replace(molecules=mole_aligned, output_shape=shape)

    def construct_landscape(
        self,
        template: TemplateInputType,
        *,
        mask: MaskInputType = None,
        max_shifts: nm | tuple[nm, nm, nm] = 1.0,
        alignment_model: type[BaseAlignmentModel] | AlignmentFactory = ZNCCAlignment,
        upsample: int = 1,
        **align_kwargs,
    ) -> da.Array:
        """
        Construct a dask array of correlation landscape.

        This method internally calls the ``landscape`` method of the input alignment
        model.

        Returns
        -------
        dask array
            If input alignment model has rotation or multiple templates, the output shape will
            be (P, M, Nz, Ny, Nx), otherwise (P, Nz, Ny, Nx), where P is for each molecules, M
            is the number of created template images, and Nz, Ny and Nx is the up-sampled shape
            of search range.
        """
        max_shifts = _normalize_max_shifts(max_shifts)
        _max_shifts_px = tuple(np.asarray(max_shifts) / self.scale)

        model = alignment_model(
            self.normalize_template(template),
            self.normalize_mask(mask),
            **align_kwargs,
        )

        if model.is_multi_templates:
            task_shape = (model.niter,) + tuple(
                2 * np.ceil(_max_shifts_px).astype(np.int32) + 1
            )
        else:
            task_shape = tuple(2 * np.ceil(_max_shifts_px).astype(np.int32) + 1)
        task_arrays = (
            self.replace(output_shape=model.input_shape)
            .iter_mapping_tasks(
                model.landscape,
                max_shifts=_max_shifts_px,
                upsample=upsample,
                var_kwarg=dict(
                    quaternion=self.molecules.quaternion(),
                    pos=self.molecules.pos / self.scale,
                ),
            )
            .tolist()
            .asarrays(shape=task_shape, dtype=np.float32)
        )
        return da.stack(task_arrays, axis=0)

    def apply(
        self,
        func: AggFunction[_R] | Iterable[AggFunction[_R]],
        *more_funcs: AggFunction[_R],
        schema: list[str] | dict[str, type[pl.DataType]] | None = None,
    ) -> pl.DataFrame:
        """
        Apply functions to subtomograms.

        Parameters
        ----------
        funcs : callable or list of callable
            Functions that take a subtomogram as input and return a scalar.
        schema : list[str], optional
            DataFrame schema.

        Returns
        -------
        pl.DataFrame
            Result table.
        """
        all_tasks: list[DaskTaskList[_R]] = []
        if more_funcs:
            func = [func] + list(more_funcs)  # type: ignore
        if _is_iterable_of_funcs(func):
            funcs = list(func)
        else:
            funcs: list[AggFunction] = [func]  # type: ignore
        if schema is None:
            schema = [fn.__name__ for fn in funcs]
        if isinstance(schema, list):
            if len(set(schema)) != len(schema):
                raise ValueError("Schema names must be unique.")
        elif isinstance(schema, dict):
            if len(schema) != len(funcs):
                raise ValueError("Schema must have the same length as the functions.")
        else:
            raise TypeError("Schema must be a list or a dict.")
        if isinstance(self.output_shape, Unset):
            raise ValueError("Output shape is unknown.")
        for fn in funcs:
            tasks = self.construct_mapping_tasks(fn, output_shape=self.output_shape)
            all_tasks.append(tasks)
        all_results = compute(all_tasks)
        df_input = [np.array(r) for r in all_results]
        return pl.DataFrame(df_input, schema=schema)

    def fsc(
        self,
        mask: TemplateInputType | None = None,
        seed: int | None = 0,
        n_set: int = 1,
        dfreq: float = 0.05,
    ) -> pl.DataFrame:
        """
        Calculate Fourier shell correlation.

        Parameters
        ----------
        mask : np.ndarray or ImageProvider, optional
            Mask image
        seed : random seed, default is 0
            Random seed used to split subtomograms.
        n_set : int, default is 1
            Number of split set of averaged images.
        dfreq : float, optional
            Frequency sampling width. Automatically determined if not provided.

        Returns
        -------
        pl.DataFrame
            A data frame with FSC results.

        See Also
        --------
        fsc_with_average
        """
        return self.fsc_with_average(mask, seed, n_set, dfreq)[0]

    def fsc_with_average(
        self,
        mask: TemplateInputType | None = None,
        seed: int | None = 0,
        n_set: int = 1,
        dfreq: float | None = None,
    ) -> tuple[pl.DataFrame, NDArray[np.float32]]:
        """
        Calculate Fourier shell correlation and the resulting averaged image.

        Parameters
        ----------
        mask : np.ndarray or ImageProvider, optional
            Mask image
        seed : random seed, default is 0
            Random seed used to split subtomograms.
        n_set : int, default is 1
            Number of split set of averaged images.
        dfreq : float, optional
            Frequency sampling width. Automatically determined if not provided.

        Returns
        -------
        pl.DataFrame and 3D array
            A data frame with FSC results and the average image. Data frame has
            columns "freq", "FSC-0", ..., "FSC-{n-1}" where n is the number of
            split.

        See Also
        --------
        fsc
        """
        if mask is None:
            _mask = 1.0
            output_shape = self.output_shape
            if isinstance(output_shape, Unset):
                raise TypeError("Output shape is unknown.")
        elif isinstance(mask, ImageProvider):
            _mask = mask(self.scale)
            output_shape = _mask.shape
        else:
            _mask = mask
            output_shape = mask.shape

        if n_set <= 0:
            raise ValueError("'n_set' must be positive.")
        dfq = 1.5 / min(output_shape) if dfreq is None else dfreq
        img = self.average_split(
            n_set=n_set,
            seed=seed,
            squeeze=False,
            output_shape=output_shape,
        )
        img[:] -= img.mean()  # normalize to minimize the sharp mask edge effect
        fsc_all: dict[str, np.ndarray] = {}
        freq = np.zeros(0, dtype=np.float32)
        for i in range(n_set):
            img0, img1 = img[i]
            freq, fsc = _utils.fourier_shell_correlation(
                img0 * _mask, img1 * _mask, dfreq=dfq
            )
            fsc_all[f"FSC-{i}"] = fsc

        out: dict[str, NDArray[np.float32]] = {"freq": freq}
        out.update(fsc_all)
        return pl.DataFrame(out), np.mean(img[0], axis=0)

    def classify(
        self,
        template: TemplateInputType | None = None,
        mask: MaskInputType = None,
        *,
        cutoff: float = 0.5,
        n_components: int = 2,
        n_clusters: int = 2,
        tilt: tuple[float, float] | None = None,
        tilt_range=None,
        seed: int = 0,
        label_name: str = "cluster",
    ) -> ClassificationResult:
        """
        Classify 3D densities by PCA of wedge-masked differences.

        Parameters
        ----------
        template : 3D array, optional
            Template image to calculate the difference. If not given, average image will
            be used.
        mask : 3D array, optional
            Soft mask of the same shape as the template or the output_shape parameter.
        n_components : int, default is 2
            Number of PCA components.
        n_clusters : int, default is 2
            Number of classes.
        tilt_range : (float, float), optional
            Tilt range in degree.
        seed : int, default is 0
            Random seed used for K-means clustering.
        label_name : str, default is "cluster"
            Column name used for the output classes.

        Returns
        -------
        ClassificationResult
            Tuple of SubtomogramLoader and PCA classifier object.

        References
        ----------
        - Heumann, J. M., Hoenger, A., & Mastronarde, D. N. (2011). Clustering and variance
          maps for cryo-electron tomography using wedge-masked differences. Journal of
          structural biology, 175(3), 288-299.
        """
        from acryo.classification import PcaClassifier

        if template is not None:
            template = self.normalize_template(template)
        _mask = self.normalize_mask(mask)
        if isinstance(_mask, np.ndarray):
            shape = _mask.shape
        elif isinstance(template, np.ndarray):
            shape = template.shape
        elif isinstance(self.output_shape, Unset):
            raise ValueError("Cannot determine output shape.")
        else:
            shape = self.output_shape

        with self.cached(shape):
            if template is None:
                template = self.average(shape)
            if tilt_range is not None:
                warnings.warn(
                    "tilt_range is deprecated. Use tilt instead",
                    DeprecationWarning,
                )
                tilt = tilt_range
            model = ZNCCAlignment(template, _mask, cutoff=cutoff, tilt=tilt)

            # PCA requires aggregation along the first axis.
            # Rechunk to improve performance.
            stack = (
                self.iter_mapping_tasks(
                    model.masked_difference,
                    output_shape=shape,
                    var_kwarg=dict(quaternion=self.molecules.quaternion()),
                )
                .tolist()
                .tostack(shape=shape, dtype=np.float32)
                .rechunk(("auto",) + shape)  # type: ignore
            )

            clf = PcaClassifier(
                stack,
                model.mask,
                n_components=n_components,
                n_clusters=n_clusters,
                seed=seed,
            )
            clf.run()

        mole = self.molecules.copy()
        mole.features = mole.features.with_columns(pl.Series(label_name, clf._labels))
        new = self.replace(molecules=mole)
        return ClassificationResult(new, clf)

    def reshape(
        self,
        template: TemplateInputType | None = None,
        mask: MaskInputType = None,
        shape: tuple[int, int, int] | None = None,
    ) -> Self:
        """Return a new loader with appropriate output shape."""
        shapes: set[tuple[int, int, int]] = set()
        if template is not None:
            shapes.add(self.normalize_template(template).shape)
        if mask is not None:
            if isinstance(mask, np.ndarray):
                shapes.add(mask.shape)
            elif isinstance(mask, ImageProvider):
                shapes.add(mask(self.scale).shape)
        if shape is not None:
            shapes.add(shape)
        if len(shapes) == 0:
            raise ValueError("Cannot infer shape from the input.")
        elif len(shapes) > 1:
            raise ValueError("Inconsistent shapes.")
        return self.replace(output_shape=shapes.pop())

    def head(self, n: int = 10) -> Self:
        """Return a new loader with the first n molecules."""
        return self.replace(molecules=self.molecules.head(n))

    def tail(self, n: int = 10) -> Self:
        """Return a new loader with the last n molecules."""
        return self.replace(molecules=self.molecules.tail(n))

    def sample(self, n: int = 10, seed: int | None = None) -> Self:
        """Return a new loader with randomly sampled molecules."""
        return self.replace(molecules=self.molecules.sample(n, seed))

    def filter(
        self,
        predicate: pl.Expr | str | pl.Series | list[bool] | np.ndarray,
    ) -> Self:
        """Return a new loader with filtered molecules."""
        out = self.replace(molecules=self.molecules.filter(predicate))
        if (cached := self._get_cached_array(None, None)) is not None:
            if isinstance(predicate, (str, pl.Expr)):
                sl = self.molecules.features.select(predicate).to_numpy().ravel()
            else:
                sl = np.asarray(predicate)
            CACHE.cache_array(cached[sl], id(out))
        return out

    @overload
    def groupby(self, by: IntoExpr) -> LoaderGroup[str, Self]:
        ...

    @overload
    def groupby(
        self, by: Sequence[IntoExpr] | tuple[IntoExpr, ...]
    ) -> LoaderGroup[tuple[str, ...], Self]:
        ...

    def groupby(self, by):
        """
        Group loader by given feature column(s).

        >>> for key, loader in loader.groupby("score"):
        ...     print(key, loader)
        """
        if isinstance(by, (str, pl.Expr)):
            _by = by
        else:
            _by = tuple(by)
        return LoaderGroup._from_loader(self, _by)

    group_by = groupby  # alias

    def _get_output_shape(self, output_shape: _ShapeType) -> tuple[pixel, ...]:
        """Normalize the `output_shape` parameter."""
        if output_shape is None:
            if isinstance(self.output_shape, Unset):
                raise ValueError("Output shape is unknown.")
            _output_shape = self.output_shape
        else:
            _output_shape = _misc.normalize_shape(output_shape, ndim=3)
        return _output_shape

    def normalize_template(self, template: TemplateInputType) -> NDArray[np.float32]:
        """Resolve any template input type to a 3D array."""
        if isinstance(template, np.ndarray):
            return template
        elif isinstance(template, ImageProvider):
            return template(self.scale)
        raise TypeError(f"Invalid template type: {type(template)}")

    def normalize_mask(self, mask: MaskInputType) -> MaskType:
        """Resolve any mask input type to a 3D array."""
        if isinstance(mask, np.ndarray) or mask is None:
            return mask
        elif isinstance(mask, ImageProvider):
            out = mask(self.scale)
            return np.asarray(out, dtype=np.float32)
        elif isinstance(mask, ImageConverter):
            return mask.with_scale(self.scale)
        raise TypeError(f"Invalid mask type: {type(mask)}")

    @overload
    def normalize_input(
        self,
        template: None = None,
        mask: MaskInputType = None,
    ) -> tuple[None, NDArray[np.float32] | None]:
        ...

    @overload
    def normalize_input(
        self,
        template: TemplateInputType,
        mask: MaskInputType = None,
    ) -> tuple[NDArray[np.float32], NDArray[np.float32] | None]:
        ...

    def normalize_input(self, template=None, mask=None):
        """Resolve any template and mask input types to 3D arrays."""
        if template is not None:
            _template = self.normalize_template(template)
            _mask = self.normalize_mask(mask)
            if callable(_mask):
                _mask = _mask(_template)
            return _template, _mask
        else:
            _mask = self.normalize_mask(mask)
            if callable(mask):
                raise TypeError("Cannot determine mask array without template.")
            return None, _mask


class ClassificationResult(NamedTuple):
    """Tuple of classification results."""

    loader: LoaderBase
    classifier: PcaClassifier


def check_input(
    order: int,
    output_shape: pixel | tuple[pixel, pixel, pixel] | Unset,
    scale: float,
    corner_safe: bool,
    ndim: int,
):
    # check interpolation order
    if order not in (0, 1, 3):
        raise ValueError(
            f"The third argument 'order' must be 0, 1 or 3, got {order!r}."
        )

    # check output_shape
    if isinstance(output_shape, Unset):
        _output_shape = output_shape
    elif output_shape is None:
        _output_shape = Unset()
    else:
        _output_shape = _misc.normalize_shape(output_shape, ndim=ndim)

    # check scale
    _scale = float(scale)
    if _scale <= 0:
        raise ValueError("Negative scale is not allowed.")

    _corner_safe = bool(corner_safe)
    return order, _output_shape, _scale, _corner_safe


def _is_iterable_of_funcs(x: Any) -> TypeGuard[Iterable[AggFunction]]:
    if not hasattr(x, "__iter__"):
        return False
    return all(callable(f) for f in x)


def _normalize_max_shifts(x: nm | tuple[nm, nm, nm]) -> tuple[nm, nm, nm]:
    if hasattr(x, "__iter__"):
        tup = tuple(float(x0) for x0 in x)  # type: ignore
        if len(tup) != 3:
            raise ValueError(
                "max_shifts must be a 3-tuple if multiple values are given."
            )
        return tup
    return (float(x),) * 3  # type: ignore



# pyright: reportPrivateImportUsage=false

from __future__ import annotations
from types import MappingProxyType

import weakref
from typing import Hashable, Iterable, Iterator, TYPE_CHECKING
import numpy as np
import polars as pl

from dask import array as da

from acryo import _utils
from acryo.molecules import Molecules
from acryo.backend import Backend
from acryo._types import nm, pixel
from acryo._dask import DaskArrayList
from acryo.loader._base import LoaderBase
from acryo.loader._loader import SubtomogramLoader, Unset

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from typing_extensions import Self
    from acryo.loader._base import _ShapeType

IMAGE_ID_LABEL = "image-id"


class BatchLoader(LoaderBase):
    """
    Collection of tomograms and their molecules.

    A `BatchLoader` is similar to a list of `SubtomogramLoader` objects, but with
    better consistency in loader parameters and more convenient access to the molecules.
    This class is useful for processing many tomograms at once, especially when you want
    to get the average of all the molecules available.
    """

    def __init__(
        self,
        order: int = 3,
        scale: nm = 1.0,
        output_shape: pixel | tuple[pixel, pixel, pixel] | Unset = Unset(),
        corner_safe: bool = False,
    ) -> None:
        super().__init__(order, scale, output_shape, corner_safe)
        self._images: dict[Hashable, NDArray[np.float32] | da.Array] = {}
        self._molecules: Molecules = Molecules.empty([IMAGE_ID_LABEL])

    def __repr__(self) -> str:
        loaders_repr = []
        for loader in self.loaders:
            shape = loader.image.shape
            mole_repr = repr(loader.molecules)
            loaders_repr.append(
                f"{loader.__class__.__name__}(tomogram={shape}, "
                f"molecules={mole_repr})"
            )
        loaders_repr = ", ".join(loaders_repr)
        return (
            f"{self.__class__.__name__}(loaders=[{loaders_repr}], "
            f"output_shape={self.output_shape}, order={self.order}, "
            f"scale={self.scale:.4f})"
        )

    @property
    def loaders(self) -> LoaderAccessor:
        """Interface to access the subtomogram loaders."""
        return LoaderAccessor(self)

    @property
    def images(self) -> MappingProxyType:
        """All the images in the collection."""
        return MappingProxyType(self._images)

    def add_tomogram(
        self,
        image: np.ndarray | da.Array,
        molecules: Molecules,
        image_id: Hashable | None = None,
    ) -> Self:
        """
        Add a tomogram and its molecules to the collection.

        Parameters
        ----------
        image : np.ndarray or da.Array
            Tomogram image. This argument is passed directly to the `SubtomogramLoader`
            constructor.
        molecules : Molecules
            Molecules in the tomogram (corresponding to the ``image`` argument). This
            argument is passed directly to the `SubtomogramLoader` constructor.
        image_id : Hashable, optional
            Identifier for the tomogram. If not provided, a unique identifier will be
            generated. This identifier is used to tag molecules with the tomogram they
            reside in.
        """
        if image_id is None:
            image_id = len(self._images)
            while image_id in self._images:
                image_id += 1
        molecules = molecules.copy()
        molecules.features = molecules.features.with_columns(
            pl.Series(IMAGE_ID_LABEL, np.full((len(molecules)), image_id))
        )
        _molecules_new = self._molecules.concat_with(molecules)

        self._images[image_id] = image
        self._molecules = _molecules_new
        return self

    def add_loader(self, loader: SubtomogramLoader | BatchLoader) -> Self:
        """Add a subtomogram loader or a batch loader to the collection."""
        if isinstance(loader, SubtomogramLoader):
            image = loader.image
            molecules = loader.molecules
            self.add_tomogram(image, molecules)
        elif isinstance(loader, BatchLoader):
            for sub in loader.loaders:
                self.add_loader(sub)
        else:
            raise TypeError(f"Cannot add {type(loader)}.")
        return self

    @classmethod
    def from_loaders(
        cls,
        loaders: Iterable[SubtomogramLoader],
        order: int = 3,
        scale: nm = 1.0,
        output_shape: pixel | tuple[pixel, pixel, pixel] | Unset = Unset(),
        corner_safe: bool = False,
    ) -> Self:
        """Construct a loader from a list of loaders."""
        self = cls(
            order=order,
            scale=scale,
            output_shape=output_shape,
            corner_safe=corner_safe,
        )
        for loader in loaders:
            self.add_loader(loader)
        return self

    @property
    def molecules(self) -> Molecules:
        """Collect all the molecules from all the loaders"""
        return self._molecules

    def replace(
        self,
        molecules: Molecules | None = None,
        output_shape: _ShapeType | Unset = None,
        order: int | None = None,
        scale: float | None = None,
        corner_safe: bool | None = None,
    ) -> Self:
        """Return a new instance with different parameter(s)."""
        if output_shape is None:
            output_shape = self.output_shape
        if order is None:
            order = self.order
        if scale is None:
            scale = self.scale
        if corner_safe is None:
            corner_safe = self.corner_safe
        out = self.__class__(
            order=order,
            scale=scale,
            output_shape=output_shape,
            corner_safe=corner_safe,
        )
        out._images = self._images.copy()
        if molecules is None:
            out._molecules = self._molecules.copy()
        else:
            out._molecules = molecules
            _id_exists = set(molecules.features[IMAGE_ID_LABEL].unique())
            for k in list(out._images.keys()):
                if k not in _id_exists:
                    out._images.pop(k)
        return out

    def binning(self, binsize: int, *, compute: bool = False) -> Self:
        """Return a new instance with binned images."""
        if binsize == 1:
            return self.copy()
        tr = -(binsize - 1) / 2 * self.scale
        molecules = self.molecules.translate([tr, tr, tr])
        _images: dict[Hashable, NDArray[np.float32] | da.Array] = {}
        _compute_dict: dict[Hashable, da.Array] = {}
        for _id, image in self._images.items():
            binned_image = _utils.bin_image(image, binsize=binsize)
            if isinstance(binned_image, da.Array) and compute:
                _compute_dict[_id] = binned_image
            _images[_id] = binned_image

        if _compute_dict:
            _computed = da.compute(_compute_dict)
            _images.update(_computed)

        out = self.replace(
            molecules=molecules,
            scale=self.scale * binsize,
        )
        out._images = _images
        return out

    def construct_loading_tasks(
        self,
        output_shape: _ShapeType = None,
        backend: Backend | None = None,
    ) -> DaskArrayList:
        """Construct batch loading tasks."""
        _backend = backend or Backend()
        return DaskArrayList.concat(
            loader.construct_loading_tasks(output_shape=output_shape, backend=_backend)
            for loader in self.loaders
        )


class LoaderAccessor:
    def __init__(self, collection: BatchLoader):
        self._loader = weakref.ref(collection)

    def __getitem__(self, idx: int) -> SubtomogramLoader:
        ldr = self._get_loader()
        mole = ldr.filter(pl.col(IMAGE_ID_LABEL) == idx).molecules
        if mole.features.shape[0] == 0:
            raise KeyError(idx)
        image = ldr._images[idx]
        loader = SubtomogramLoader(
            image,
            mole,
            ldr.order,
            ldr.scale,
            ldr.output_shape,
            ldr.corner_safe,
        )
        return loader

    def __iter__(self) -> Iterator[SubtomogramLoader]:
        ldr = self._get_loader()
        for key, group in ldr.molecules.groupby(IMAGE_ID_LABEL):
            image = ldr._images[key]
            loader = SubtomogramLoader(
                image,
                group,
                ldr.order,
                ldr.scale,
                ldr.output_shape,
                ldr.corner_safe,
            )
            yield loader

    def __len__(self) -> int:
        """Number of loaders."""
        ldr = self._get_loader()
        return len(ldr._images)

    def _get_loader(self) -> BatchLoader:
        loader = self._loader()
        if loader is None:
            raise RuntimeError("Loader is no longer available.")
        return loader



from __future__ import annotations

from typing import Callable, overload, TypeVar, Generic
from typing_extensions import Self

import numpy as np
from numpy.typing import NDArray
from acryo._types import nm

_R = TypeVar("_R", np.ndarray, "list[np.ndarray]")


class _Pipeline:
    def __init__(self, fn: Callable):
        self._func = fn
        self.__name__ = getattr(fn, "__name__", repr(fn))

    def with_name(self, name: str) -> Self:
        self.__name__ = name
        return self

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.__name__}>"

    def __radd__(self, other) -> Self:
        return self + other

    def __rsub__(self, other) -> Self:
        return self - other

    def __rmul__(self, other) -> Self:
        return self * other

    def __rtruediv__(self, other) -> Self:
        return self / other


class ImageProvider(_Pipeline, Generic[_R]):
    """Function that provides an image at a given scale."""

    def __init__(self, provider: Callable[[nm], _R]):
        super().__init__(provider)

    def provide(self, scale: nm) -> _R:
        """Provide an image at a given scale."""
        out = self._func(scale)
        assert_n_3d_arrays(out, self._func)
        return out

    def __call__(self, scale: nm) -> _R:
        return self.provide(scale)

    def __add__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(lambda scale: self(scale) + other(scale)).with_name(
                f"({self.__name__} + {other.__name__})"
            )
        return self.__class__(lambda scale: self(scale) + other)

    def __sub__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(lambda scale: self(scale) - other(scale)).with_name(
                f"({self.__name__} - {other.__name__})"
            )
        return self.__class__(lambda scale: self(scale) - other)

    def __mul__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(lambda scale: self(scale) * other(scale)).with_name(
                f"{self.__name__} * {other.__name__}"
            )
        return self.__class__(lambda scale: self(scale) * other)

    def __truediv__(self, other) -> ImageProvider:
        if np.isscalar(other) and other == 0:
            raise ZeroDivisionError("Cannot divide by zero.")
        if isinstance(other, ImageProvider):
            return self.__class__(lambda scale: self(scale) / other(scale)).with_name(
                f"{self.__name__} / {other.__name__}"
            )
        return self.__class__(lambda scale: self(scale) / other)

    def __eq__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(lambda scale: self(scale) == other(scale)).with_name(
                f"{self.__name__} == {other.__name__}"
            )
        return self.__class__(lambda scale: self(scale) == other)

    def __ne__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(lambda scale: self(scale) != other(scale)).with_name(
                f"{self.__name__} != {other.__name__}"
            )
        return self.__class__(lambda scale: self(scale) != other)

    def __lt__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(
                lambda scale: _lt(self(scale), other(scale))
            ).with_name(f"{self.__name__} < {other.__name__}")
        return self.__class__(lambda scale: self(scale) < other)

    def __le__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(
                lambda scale: _le(self(scale), other(scale))
            ).with_name(f"{self.__name__} <= {other.__name__}")
        return self.__class__(lambda scale: self(scale) <= other)

    def __gt__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(
                lambda scale: _gt(self(scale), other(scale))
            ).with_name(f"{self.__name__} > {other.__name__}")
        return self.__class__(lambda scale: self(scale) > other)

    def __ge__(self, other) -> ImageProvider:
        if isinstance(other, ImageProvider):
            return self.__class__(
                lambda scale: _ge(self(scale), other(scale))
            ).with_name(f"{self.__name__} >= {other.__name__}")
        return self.__class__(lambda scale: self(scale) >= other)

    def __neg__(self) -> ImageProvider:
        return self.__class__(lambda scale: -self(scale)).with_name(
            f"(-{self.__name__})"
        )


class ImageConverter(_Pipeline):
    """Functionthat convert an image."""

    def __init__(
        self, converter: Callable[[NDArray[np.float32], nm], NDArray[np.float32]]
    ):
        super().__init__(converter)

    def convert(self, image: NDArray[np.float32], scale: nm) -> NDArray[np.float32]:
        out = self._func(image, scale)
        _assert_3d_array(out, self._func)
        return out

    def __call__(self, image: NDArray[np.float32], scale: nm) -> NDArray[np.float32]:
        return self.convert(image, scale)

    @overload
    def compose(self, other: ImageProvider) -> ImageProvider:
        ...

    @overload
    def compose(self, other: ImageConverter) -> ImageConverter:
        ...

    def compose(self, other):
        """Function composition"""
        if isinstance(other, ImageProvider):
            fn = lambda scale: self(other(scale), scale)
        elif isinstance(other, ImageConverter):
            fn = lambda x, scale: self(other(x, scale), scale)
        else:
            raise TypeError("Cannot compose with a non-pipeline object.")
        return other.__class__(fn).with_name(f"{self.__name__}○{other.__name__}")

    __matmul__ = compose

    def with_scale(
        self, scale: nm
    ) -> Callable[[NDArray[np.float32]], NDArray[np.float32]]:
        """Partialize converter with a given scale."""

        def fn(img: NDArray[np.float32]) -> NDArray[np.float32]:
            return self(img, scale)

        fn.__name__ = self._func.__name__
        return fn

    def __add__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: self(x, scale) + other(x, scale)
            ).with_name(f"({self.__name__} + {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: self(x, scale) + other(scale)
            ).with_name(f"({self.__name__} + {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) + other)

    def __sub__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: self(x, scale) - other(x, scale)
            ).with_name(f"({self.__name__} - {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: self(x, scale) - other(scale)
            ).with_name(f"({self.__name__} - {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) - other)

    def __mul__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: self(x, scale) * other(x, scale)
            ).with_name(f"({self.__name__} * {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: self(x, scale) * other(scale)
            ).with_name(f"({self.__name__} * {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) * other)

    def __truediv__(self, other) -> ImageConverter:
        if np.isscalar(other) and other == 0:
            raise ZeroDivisionError("Cannot divide by zero.")
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: self(x, scale) / other(x, scale)
            ).with_name(f"({self.__name__} / {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: self(x, scale) / other(scale)
            ).with_name(f"({self.__name__} / {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) / other)

    def __eq__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: self(x, scale) == other(x, scale)
            ).with_name(f"({self.__name__} == {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: self(x, scale) == other(scale)
            ).with_name(f"({self.__name__} == {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) == other)

    def __ne__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: self(x, scale) != other(x, scale)
            ).with_name(f"({self.__name__} != {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: self(x, scale) != other(scale)
            ).with_name(f"({self.__name__} != {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) != other)

    def __gt__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: _gt(self(x, scale), other(x, scale))
            ).with_name(f"({self.__name__} > {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: _gt(self(x, scale), other(scale))
            ).with_name(f"({self.__name__} > {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) > other)

    def __ge__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: _ge(self(x, scale), other(x, scale))
            ).with_name(f"({self.__name__} >= {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: _ge(self(x, scale), other(scale))
            ).with_name(f"({self.__name__} >= {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) >= other)

    def __lt__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: _lt(self(x, scale), other(x, scale))
            ).with_name(f"({self.__name__} < {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: _lt(self(x, scale), other(scale))
            ).with_name(f"({self.__name__} < {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) < other)

    def __le__(self, other) -> ImageConverter:
        if isinstance(other, ImageConverter):
            return self.__class__(
                lambda x, scale: _le(self(x, scale), other(x, scale))
            ).with_name(f"({self.__name__} <= {other.__name__})")
        elif isinstance(other, ImageProvider):
            return self.__class__(
                lambda x, scale: _le(self(x, scale), other(scale))
            ).with_name(f"({self.__name__} <= {other.__name__})")
        return self.__class__(lambda x, scale: self(x, scale) <= other)

    def __neg__(self) -> ImageConverter:
        return self.__class__(lambda x, scale: -self(x, scale))


def assert_n_3d_arrays(out, func):
    if isinstance(out, np.ndarray):
        return _assert_3d_array(out, func)
    for o in out:
        _assert_3d_array(o, func)


def _assert_3d_array(out, func):
    if not isinstance(out, np.ndarray):
        raise TypeError(
            f"Function {func!r} did not return a numpy.ndarray (got {type(out)})"
        )
    if out.ndim != 3:
        raise ValueError(f"Wrong image dimensionality: {out.shape}")


def _ge(a, b) -> NDArray[np.float32]:
    return np.greater_equal(a, b, dtype=np.float32)


def _le(a, b) -> NDArray[np.float32]:
    return np.less_equal(a, b, dtype=np.float32)


def _gt(a, b) -> NDArray[np.float32]:
    return np.greater(a, b, dtype=np.float32)


def _lt(a, b) -> NDArray[np.float32]:
    return np.less(a, b, dtype=np.float32)



# pyright: reportPrivateImportUsage=false
from __future__ import annotations

from typing import (
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    TypeVar,
    TYPE_CHECKING,
    Union,
    overload,
)
import numpy as np
from numpy.typing import NDArray
import polars as pl
from dask import array as da

from acryo.alignment import (
    BaseAlignmentModel,
    ZNCCAlignment,
    AlignmentFactory,
    AlignmentResult,
)

from acryo import _utils
from acryo._types import nm
from acryo._dask import compute, DaskTaskList
from acryo.loader import _misc
from acryo.backend import Backend, AnyArray

if TYPE_CHECKING:
    from dask.delayed import Delayed
    from numpy.typing import NDArray
    from acryo.loader._base import (
        LoaderBase,
        Unset,
        _ShapeType,
        TemplateInputType,
        MaskInputType,
        AggFunction,
    )

IntoExpr = Union[str, pl.Expr]
_K = TypeVar("_K", bound=Union[str, tuple[str, ...]])
_L = TypeVar("_L", bound="LoaderBase")
_R = TypeVar("_R")


class LoaderGroup(Generic[_K, _L]):
    """A groupby-like object for subtomogram loaders."""

    def __init__(
        self,
        it: Iterable[tuple[_K, _L]],
    ):
        self._it = it

    @property
    def keys(self) -> list[_K]:
        """All the keys in the group."""
        return [key for key, _ in self._it]

    @overload
    @classmethod
    def _from_loader(cls, loader: _L, by: IntoExpr) -> LoaderGroup[str, _L]:
        ...

    @overload
    @classmethod
    def _from_loader(
        cls, loader: _L, by: tuple[IntoExpr, ...]
    ) -> LoaderGroup[tuple[str, ...], _L]:
        ...

    @classmethod
    def _from_loader(cls, loader: _L, by: IntoExpr | tuple[IntoExpr, ...]):  # type: ignore[override]
        return cls(
            LoaderGroupByIterator(
                loader,
                by,
                order=loader.order,
                scale=loader.scale,
                output_shape=loader.output_shape,
                corner_safe=loader.corner_safe,
            )
        )

    def __iter__(self) -> Iterator[tuple[_K, _L]]:
        yield from self._it

    def average(
        self,
        output_shape: _ShapeType = None,
        *,
        backend: Backend | None = None,
    ) -> ArrayDict[_K]:
        """Calculate average images."""
        xp = backend or Backend()
        tasks = []
        keys: list[_K] = []
        for key, loader in self:
            keys.append(key)
            _output_shape = loader._get_output_shape(output_shape)
            dsk = loader.construct_dask(_output_shape, backend=xp)
            tasks.append(da.mean(dsk, axis=0))

        out: list[AnyArray[np.float32]] = da.compute(tasks)[0]
        return ArrayDict({key: xp.asnumpy(img) for key, img in zip(keys, out)})

    def average_split(
        self,
        n_set: int = 1,
        seed: int | None = 0,
        squeeze: bool = True,
        output_shape: _ShapeType = None,
        *,
        backend: Backend | None = None,
    ) -> ArrayDict[_K]:
        """
        Split subtomograms into two set and average separately.

        This method executes pairwise subtomogram averaging using randomly
        selected molecules, which is useful for calculation of such as Fourier
        shell correlation.

        Parameters
        ----------
        n_set : int, default is 1
            Number of split set of averaged image.
        seed : random seed, default is 0
            Random seed to determine how subtomograms will be split.
        squeeze : bool, default is True
            If true and n_set is 1, return a dictionary of 4D arrays.
        output_shape : tuple of int, optional
            Output shape of the averaged image. If not given, the default output
            shape of the loader objects will be used.

        Returns
        -------
        dict of np.ndarray
            Averaged images with keys of the group keys.
        """
        xp = backend or Backend()
        rng = np.random.default_rng(seed=seed)

        all_tasks: list[list[da.Array]] = []
        for key, loader in self:
            output_shape = loader._get_output_shape(output_shape)
            dask_array = loader.construct_dask(output_shape=output_shape, backend=xp)
            nmole = dask_array.shape[0]
            tasks: list[da.Array] = []
            for _ in range(n_set):
                ind0, ind1 = _misc.random_splitter(rng, nmole)
                _stack = da.stack(
                    [
                        da.mean(dask_array[ind0], axis=0),
                        da.mean(dask_array[ind1], axis=0),
                    ],
                    axis=0,
                )
                tasks.append(_stack)
            all_tasks.append(tasks)

        computed: list[list[AnyArray[np.float32]]] = da.compute(all_tasks)[0]
        out = ArrayDict()
        for (key, loader), stack in zip(self, computed):
            stack_np = xp.asnumpy(xp.stack(stack, axis=0))
            if squeeze and n_set == 1:
                stack_np: NDArray[np.float32] = stack_np[0]
            out[key] = stack_np
        return out

    def align(
        self,
        template: TemplateInputType | Mapping[_K, TemplateInputType],
        *,
        mask: MaskInputType = None,
        max_shifts: nm | tuple[nm, nm, nm] = 1.0,
        alignment_model: type[BaseAlignmentModel] | AlignmentFactory = ZNCCAlignment,
        backend: Backend | None = None,
        **align_kwargs,
    ) -> LoaderGroup[_K, _L]:
        """
        Align subtomograms to the template image.

        This method conduct so called "subtomogram alignment". Only shifts and rotations
        are calculated in this method. To get averaged image, you'll have to run "average"
        method using the resulting LoaderGroup instance.

        Parameters
        ----------
        template : 3D array or ImageProvider or mapping
            Template image.
        mask : np.ndarray or callable of np.ndarray to np.ndarray optional
            Mask image. Must in the same shape as the template.
        max_shifts : int or tuple of int, default is (1., 1., 1.)
            Maximum shift between subtomograms and template.
        alignment_model : subclass of BaseAlignmentModel, optional
            Alignment model class used for subtomogram alignment. By default,
            ``ZNCCAlignment`` will be used.
        align_kwargs : optional keyword arguments
            Additional keyword arguments passed to the input alignment model.

        Returns
        -------
        LoaderGroup
            A loader group instance with updated molecules.
        """
        all_tasks: list[DaskTaskList[AlignmentResult]] = []
        template_map = _normalize_template(template)
        input_shape: tuple[int, int, int] | None = None
        for key, loader in self:
            model = alignment_model(
                loader.normalize_template(template_map[key]),
                loader.normalize_mask(mask),
                **align_kwargs,
            )
            _max_shifts_px = np.asarray(max_shifts) / loader.scale
            tasks = loader.construct_mapping_tasks(
                model.align,
                max_shifts=_max_shifts_px,
                output_shape=model.input_shape,
                backend=backend,
                var_kwarg=dict(
                    quaternion=loader.molecules.quaternion(),
                    pos=loader.molecules.pos / loader.scale,
                ),
            )
            all_tasks.append(tasks)
            input_shape = model.input_shape

        if input_shape is None:
            raise RuntimeError("LoaderGroup has no loader.")

        all_results = compute(all_tasks)
        out: list[tuple[_K, _L]] = []
        for (key, loader), results in zip(self, all_results):
            new = loader._post_align(results, input_shape)
            out.append((key, new))
        return self.__class__(out)

    def align_no_template(
        self,
        *,
        mask: MaskInputType = None,
        max_shifts: nm | tuple[nm, nm, nm] = 1.0,
        output_shape: _ShapeType = None,
        alignment_model: type[BaseAlignmentModel] = ZNCCAlignment,
        backend: Backend | None = None,
        **align_kwargs,
    ) -> LoaderGroup[_K, _L]:
        """
        Align subtomograms without template image.

        A template-free version of :func:`align`. This method first
        calculates averaged image and uses it for the alignment template. To
        avoid loading same subtomograms twice, a memory-mapped array is created
        internally (so the second subtomogram loading is faster).

        Parameters
        ----------
        mask : np.ndarray or callable, optional
            Mask image. Must in the same shape as the template.
        max_shifts : int or tuple of int, default is (1., 1., 1.)
            Maximum shift between subtomograms and template.
        alignment_model : subclass of BaseAlignmentModel, optional
            Alignment model class used for subtomogram alignment. By default,
            ``ZNCCAlignment`` will be used.
        align_kwargs : optional keyword arguments
            Additional keyword arguments passed to the input alignment model.

        Returns
        -------
        LoaderGroup
            A loader group with updated molecules.
        """
        xp = backend or Backend()
        avg = self.average(output_shape=output_shape, backend=xp)
        return self.align(
            avg,
            mask=mask,
            max_shifts=max_shifts,
            alignment_model=alignment_model,
            backend=xp,
            **align_kwargs,
        )

    def align_multi_templates(
        self,
        templates: list[TemplateInputType] | Mapping[_K, list[TemplateInputType]],
        *,
        mask: MaskInputType = None,
        max_shifts: nm | tuple[nm, nm, nm] = 1.0,
        alignment_model: type[BaseAlignmentModel] = ZNCCAlignment,
        label_name: str = "labels",
        **align_kwargs,
    ) -> LoaderGroup[_K, _L]:
        """
        Align subtomograms with multiple template images.

        A multi-template version of :func:`align`. This method calculate cross
        correlation for every template and uses the best local shift, rotation and
        template.

        Parameters
        ----------
        templates: list of 3D arrays or ImageProvider or mapping
            Template images. If ImageProvider is given, the image will be provided
            accordingly using the scale of the loader object.
        mask : np.ndarray, optional
            Mask image. Must in the same shape as the template.
        max_shifts : int or tuple of int, default is (1., 1., 1.)
            Maximum shift between subtomograms and template.
        alignment_model : subclass of BaseAlignmentModel, optional
            Alignment model class used for subtomogram alignment. By default,
            ``ZNCCAlignment`` will be used.
        align_kwargs : optional keyword arguments
            Additional keyword arguments passed to the input alignment model.

        Returns
        -------
        LoaderGroup
            A loader group with updated molecules.
        """
        all_tasks: list[DaskTaskList[AlignmentResult]] = []
        template_map = _normalize_template(templates)
        input_shape: tuple[int, int, int] | None = None
        has_rotation = False
        for key, loader in self:
            _tmps = template_map[key]
            model = alignment_model(
                template=[loader.normalize_template(t) for t in _tmps],
                mask=loader.normalize_mask(mask),
                **align_kwargs,
            )
            _max_shifts_px = np.asarray(max_shifts) / loader.scale
            tasks = loader.construct_mapping_tasks(
                model.align,
                max_shifts=_max_shifts_px,
                output_shape=model.input_shape,
                var_kwarg=dict(
                    quaternion=loader.molecules.quaternion(),
                    pos=loader.molecules.pos / loader.scale,
                ),
            )
            all_tasks.append(tasks)
            input_shape = model.input_shape
            has_rotation = model.has_rotation

        if input_shape is None:
            raise RuntimeError("LoaderGroup has no loader.")

        all_results = compute(all_tasks)
        if has_rotation:
            remainder = len(templates)
        else:
            remainder = -1
        out: list[tuple[_K, _L]] = []
        for (key, loader), results in zip(self, all_results):
            new = loader._post_align_multi_templates(
                results,
                input_shape,
                remainder,
                label_name,
            )
            out.append((key, new))
        return self.__class__(out)

    def apply(
        self,
        funcs: AggFunction[_R] | Sequence[AggFunction[_R]],
        schema: list[str] | None = None,
    ) -> DataFrameDict[_K]:
        """
        Apply functions to subtomograms for each group.

        Parameters
        ----------
        funcs : callable or list of callable
            Functions that take a subtomogram as input and return a scalar.
        schema : list[str], optional
            DataFrame schema.

        Returns
        -------
        dict of pl.DataFrame
            Result tables for each group.
        """
        all_tasks: list[list[Delayed]] = []
        if not isinstance(funcs, Sequence):
            _funcs = [funcs]
        else:
            _funcs = funcs
        if schema is None:
            schema = [fn.__name__ for fn in _funcs]
        if len(set(schema)) != len(schema):
            raise ValueError("Schema names must be unique.")
        keys: list[_K] = []
        xp = Backend()
        for key, loader in self:
            output_shape = loader.output_shape
            if not isinstance(output_shape, tuple):
                raise ValueError("LoaderGroup has no output shape.")
            taskset = []
            for fn in _funcs:
                f_as_np = lambda ar: fn(xp.asnumpy(ar))
                tasks = loader.construct_mapping_tasks(
                    f_as_np, output_shape=output_shape
                )
                taskset.append(list(tasks))
            all_tasks.append(taskset)
            keys.append(key)
        all_results = da.compute(all_tasks)[0]
        out = DataFrameDict()
        for key, result in zip(keys, all_results):
            out[key] = pl.DataFrame(np.array(result), schema=schema)
        return out

    def fsc(
        self,
        mask: TemplateInputType | None = None,
        seed: int | None = 0,
        n_set: int = 1,
        dfreq: float = 0.05,
    ) -> DataFrameDict[_K]:
        """
        Calculate Fourier shell correlation for each group.

        Parameters
        ----------
        mask : np.ndarray, optional
            Mask image
        seed : random seed, default is 0
            Random seed used to split subtomograms.
        n_set : int, default is 1
            Number of split set of averaged images.
        dfreq : float, default is 0.05
            Frequency sampling width.

        Returns
        -------
        pl.DataFrame
            A data frame with FSC results.
        """
        if mask is None:
            _mask = 1
            output_shape = None
        elif isinstance(mask, np.ndarray):
            _mask = mask
            output_shape = mask.shape
        else:
            _mask = 1
            output_shape = None

        if n_set <= 0:
            raise ValueError("'n_set' must be positive.")

        imgs = self.average_split(
            n_set=n_set,
            seed=seed,
            squeeze=False,
            output_shape=output_shape,
        )
        out = DataFrameDict()
        for key, img in imgs.items():
            fsc_all: dict[str, np.ndarray] = {}
            freq = np.zeros(0, dtype=np.float32)
            for i in range(n_set):
                img0, img1 = img[i]
                freq, fsc = _utils.fourier_shell_correlation(
                    img0 * _mask, img1 * _mask, dfreq=dfreq
                )
                fsc_all[f"FSC-{i}"] = fsc

            fsc_dict: dict[str, NDArray[np.float32]] = {"freq": freq}
            fsc_dict.update(fsc_all)
            out[key] = pl.DataFrame(fsc_dict)
        return out

    def count(self) -> dict[_K, int]:
        """Dictionary of the molecule count in each group."""
        return {key: loader.count() for key, loader in self}

    def filter(
        self,
        predicate: pl.Expr | str | pl.Series | list[bool] | np.ndarray,
    ) -> LoaderGroup[_K, _L]:
        """Filter the molecules in each group."""
        return self.__class__((key, loader.filter(predicate)) for key, loader in self)

    def head(self, n: int = 10) -> LoaderGroup[_K, _L]:
        """Get the first n molecules in each group."""
        return self.__class__((key, loader.head(n)) for key, loader in self)

    def tail(self, n: int = 10) -> LoaderGroup[_K, _L]:
        """Get the last n molecules in each group."""
        return self.__class__((key, loader.tail(n)) for key, loader in self)

    def sample(self, n: int = 10, seed: int | None = None) -> LoaderGroup[_K, _L]:
        """Get n random molecules in each group."""
        return self.__class__((key, loader.sample(n, seed)) for key, loader in self)


class LoaderGroupByIterator(Iterable[tuple[_K, _L]]):
    """Iterator for the loader groupby."""

    def __init__(
        self,
        loader: _L,
        by: IntoExpr | Sequence[IntoExpr],
        order: int,
        scale: float,
        output_shape: tuple[int, int, int] | Unset | None,
        corner_safe: bool,
    ):
        self._loader = loader
        self._by = by
        self._order = order
        self._scale = scale
        self._output_shape = output_shape
        self._corner_safe = corner_safe

    def __iter__(self) -> Iterator[tuple[_K, _L]]:
        loader = self._loader
        cached = loader._get_cached_array(None, None)
        index_col_name = ".index"
        if index_col_name in loader.molecules.features:
            index_col_name += "."
        index = pl.Series(index_col_name, np.arange(loader.count()))
        for key, mole in loader.molecules.with_features(index).groupby(self._by):
            _loader = loader.replace(
                molecules=mole.drop_features(index_col_name),
                order=self._order,
                scale=self._scale,
                output_shape=self._output_shape,
                corner_safe=self._corner_safe,
            )
            if cached is not None:
                sl = mole.features[index_col_name].to_numpy()
                loader._CACHE.cache_array(cached[sl], id(_loader))
            yield key, _loader  # type: ignore


_T = TypeVar("_T")


def _normalize_template(template: _T | Mapping[_K, _T]) -> Mapping[_K, _T]:
    if isinstance(template, Mapping):
        return template
    from collections import defaultdict

    return defaultdict(lambda: template)


class ArrayDict(dict[_K, NDArray[np.float32]]):
    def value_list(self) -> list[NDArray[np.float32]]:
        """Return values as a list."""
        return list(self.values())

    def value_stack(self, axis: int = 0) -> NDArray[np.float32]:
        """Return values as a stack."""
        return np.stack(self.value_list(), axis=axis)


class DataFrameDict(dict[_K, pl.DataFrame]):
    def value_list(self) -> list[pl.DataFrame]:
        """Return values as a list."""
        return list(self.values())

    def value_melt(self, key_label: str = "group") -> pl.DataFrame:
        """Return values as a melted data frame."""
        df_melt = []
        for key, df in self.items():
            df_melt.append(
                df.with_columns(pl.repeat(str(key), pl.count()).alias(key_label))
            )
        return pl.concat(df_melt)



# pyright: reportPrivateImportUsage=false
from __future__ import annotations

from typing import Any, Sequence, SupportsIndex
from abc import ABC, abstractmethod

import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation
from dask import array as da

from acryo.pipe._classes import ImageProvider
from acryo.molecules import Molecules
from acryo._rotation import normalize_rotations
from acryo._utils import compose_matrices, missing_wedge_mask
from acryo._types import nm, RotationType
from acryo._typed_scipy import spline_filter, affine_transform
from acryo._dask import DaskTaskPool


class BasePickerModel(ABC):
    def pick_molecules(
        self,
        image: da.Array,
        scale: nm = 1.0,
        **kwargs,
    ) -> Molecules:
        params, depth = self.get_params_and_depth(scale)
        if image.dtype != np.float32:
            image = image.astype(np.float32)
        task: da.Array = image.map_overlap(
            self._pick_in_chunk_wrapped,
            **params,
            **kwargs,
            # dask parameters
            depth=depth,
            trim=False,
            boundary="nearest",
            dtype=object,
            meta=np.array([]),
        )
        boxes: Sequence[MoleculesBox] = task.compute().ravel()
        mole = Molecules.concat([box.to_molecules() for box in boxes])
        mole._pos *= scale
        return mole

    def _pick_in_chunk_wrapped(
        self,
        image: NDArray[np.float32],
        block_info: dict,
        **kwargs,
    ) -> NDArray[np.object_]:
        pos, quats, features = self.pick_in_chunk(image, **kwargs)

        locs: list[tuple[int, int]] = block_info[None]["array-location"]
        for i, (start, _) in enumerate(locs):
            pos[:, i] += start

        return np.array([[[MoleculesBox(pos, quats, features)]]], dtype=object)

    @abstractmethod
    def pick_in_chunk(
        self, image: NDArray[np.float32], **kwargs
    ) -> tuple[NDArray[np.float32], NDArray[np.uint16], Any]:
        ...

    @abstractmethod
    def get_params_and_depth(
        self, scale: nm
    ) -> tuple[dict[str, Any], int | Sequence[SupportsIndex]]:
        ...


class BaseTemplateMatcher(BasePickerModel):
    """Particle picking based on parallel template matching."""

    def __init__(
        self,
        template: NDArray[np.float32] | ImageProvider,
        rotation: RotationType,
        tilt_range: tuple[float, float] = (-60, 60),
        order: int = 1,
    ) -> None:
        self._order = order
        self._quaternions = normalize_rotations(rotation)
        self._tilt_range = tilt_range
        self._template = template

    @property
    def order(self) -> int:
        """Interpolation order."""
        return self._order

    def get_params_and_depth(self, scale: nm):
        if isinstance(self._template, ImageProvider):
            template = self._template(scale)
        else:
            template = self._template

        rotators = [Rotation.from_quat(r).inv() for r in self._quaternions]
        _center = np.array(template.shape) / 2 - 0.5
        matrices = compose_matrices(_center, rotators)

        if self.order > 2:
            template = spline_filter(
                template,
                order=self.order,
                output=np.float32,
                mode="constant",
            )

        pool = DaskTaskPool.from_func(affine_transform)
        for mtx in matrices:
            pool.add_task(template, mtx, order=self.order, prefilter=False)
        mask = missing_wedge_mask(
            Rotation.from_quat([0, 0, 0, 1]), self._tilt_range, template.shape
        )
        out = pool.compute()  # rotated templates
        templates = [o * mask for o in out]
        depth = tuple(np.ceil(np.array(templates[0].shape) / 2).astype(np.uint16))
        return {"templates": templates}, depth

    def _index_to_quaternions(self, argmax_indices):
        return np.take_along_axis(
            self._quaternions, argmax_indices[:, np.newaxis], axis=0
        )

    @abstractmethod
    def pick_in_chunk(
        self, image: NDArray[np.float32], templates: list[NDArray[np.float32]], **kwargs
    ) -> tuple[NDArray[np.float32], NDArray[np.float32], Any]:
        ...


# NOTE: np.array(mole) is dangerous.
class MoleculesBox:
    def __init__(self, pos, quats, features) -> None:
        self._pos = pos
        self._quats = quats
        self._features = features

    def to_molecules(self) -> Molecules:
        return Molecules.from_quat(self._pos, self._quats, features=self._features)



from __future__ import annotations
from typing import TypeVar
import math
import numpy as np
from numpy.typing import NDArray

from acryo.backend import Backend, AnyArray

# Modified from skimage/registration/_phase_cross_correlation.py


def pcc_landscape(
    f0: AnyArray[np.complex64],
    f1: AnyArray[np.complex64],
    max_shifts: tuple[float, ...],
    backend: Backend,
):
    product = f0 * f1.conj()
    power = _abs2(backend.ifftn(product))
    power = backend.fftshift(power)
    centers = tuple(s // 2 for s in power.shape)
    slices = tuple(
        slice(max(c - int(shiftl), 0), min(c + int(shiftr) + 1, s), None)
        for c, shiftl, shiftr, s in zip(centers, max_shifts, max_shifts, power.shape)
    )
    power = power[slices]
    return power


def subpixel_pcc(
    f0: AnyArray[np.complex64],
    f1: AnyArray[np.complex64],
    upsample_factor: int,
    max_shifts: tuple[float, ...] | NDArray[np.number],
    backend: Backend,
) -> tuple[NDArray[np.float32], float]:
    product = f0 * f1.conj()
    power = _abs2(backend.ifftn(product))
    _max_shifts = np.asarray(max_shifts, dtype=np.float32)
    _int_shifts = _max_shifts.astype(np.int32)
    power = crop_by_max_shifts(power, _int_shifts, _int_shifts, backend)

    maxima = backend.unravel_index(backend.argmax(power), power.shape)
    midpoints = np.array(
        [np.fix(axis_size / 2) for axis_size in power.shape], dtype=np.float32
    )

    shifts = backend.asnumpy(maxima).astype(np.float32)
    sl = shifts > midpoints
    shifts[sl] -= np.array(power.shape, dtype=np.float32)[sl]
    # Initial shift estimate in upsampled grid
    shifts = np.fix(shifts * upsample_factor) / upsample_factor
    if upsample_factor > 1:
        upsampled_region_size = math.ceil(upsample_factor * 1.5)
        # Center of output array at dftshift + 1
        dftshift = float(np.fix(upsampled_region_size / 2.0))
        # Matrix multiply DFT around the current shift estimate
        sample_region_offset = dftshift - shifts * upsample_factor
        # Locate maximum and map back to original pixel grid
        power = _abs2(
            _upsampled_dft(
                product.conj(),
                upsampled_region_size,
                upsample_factor,
                sample_region_offset,
                backend,
            )
        )

        _lshift = (shifts + _max_shifts) * upsample_factor
        _rshift = (_max_shifts - shifts) * upsample_factor
        power = crop_by_max_shifts(
            power, _lshift.astype(np.int32), _rshift.astype(np.int32), backend
        )

        maxima = (
            backend.asnumpy(
                backend.unravel_index(backend.argmax(power), power.shape)
            ).astype(np.float32)
            - dftshift
        )
        shifts = shifts + maxima / upsample_factor
        pcc = math.sqrt(backend.asnumpy(power[tuple(int(round(m)) for m in maxima)]))
    else:
        pcc = math.sqrt(backend.asnumpy(power[maxima]))
    return shifts, pcc


_DType = TypeVar("_DType", bound=np.number)


def _upsampled_dft(
    data: AnyArray[_DType],
    upsampled_region_size: int,
    upsample_factor: int,
    axis_offsets: NDArray[np.float32],
    backend: Backend,
) -> AnyArray[_DType]:
    # if people pass in an integer, expand it to a list of equal-sized sections
    upsampled_region_sizes = [upsampled_region_size] * data.ndim

    dim_properties = list(zip(data.shape, upsampled_region_sizes, axis_offsets))

    for n_items, ups_size, ax_offset in dim_properties[::-1]:
        freq = backend.fftfreq(n_items, upsample_factor).astype(np.float32)
        nth = backend.arange(ups_size, dtype=np.float32) - ax_offset
        kernel = backend.exp(-2j * np.pi * (nth[:, np.newaxis] * freq))  # type: ignore
        data = backend.tensordot(kernel, data, axes=(1, -1))
    return data


def _abs2(a: AnyArray[np.complex64]) -> AnyArray[np.float32]:
    return a.real**2 + a.imag**2


def crop_by_max_shifts(
    power: AnyArray[_DType],
    left: NDArray[np.intp],
    right: NDArray[np.intp],
    backend: Backend,
) -> AnyArray[_DType]:
    shifted_power = backend.fftshift(power)
    centers = tuple(s // 2 for s in power.shape)
    slices = tuple(
        slice(max(c - int(shiftl), 0), min(c + int(shiftr) + 1, s), None)  # type: ignore
        for c, shiftl, shiftr, s in zip(centers, left, right, power.shape)
    )
    return backend.ifftshift(shifted_power[slices])



from __future__ import annotations

from typing import overload, Literal
from acryo.tilt._single import SingleAxisY, SingleAxisX, SingleAxis
from acryo.tilt._base import UnionAxes, NoWedge


@overload
def single_axis(
    tilt_range: tuple[float, float], axis: Literal["x", "y"] = "y"
) -> SingleAxis:
    ...


@overload
def single_axis(tilt_range: None, axis: Literal["x", "y"] = "y") -> NoWedge:
    ...


def single_axis(
    tilt_range: tuple[float, float] | None = None,
    axis: Literal["x", "y"] = "y",
) -> SingleAxis | NoWedge:
    """
    Create a single-axis missing wedge model.

    Parameters
    ----------
    tilt_range : (float, float)
        Minimum and maximum tilt angles in degrees.
    axis : str, default is "y"
        The rotation axis of tomography.
    """
    if tilt_range is None:
        return NoWedge()
    if axis == "y":
        return SingleAxisY(tilt_range)
    elif axis == "x":
        return SingleAxisX(tilt_range)
    else:
        raise ValueError(f"Invalid axis: {axis}")


def dual_axis(
    tilt_range_y: tuple[float, float],
    tilt_range_x: tuple[float, float],
) -> UnionAxes:
    """
    Create a dual-axis missing wedge model.

    Parameters
    ----------
    tilt_range_y : (float, float)
        Minimum and maximum tilt angles of the y-axis tilt series in degrees.
    tilt_range_x : (float, float)
        Minimum and maximum tilt angles of the x-axis tilt series in degrees.
    """
    return UnionAxes(
        [
            SingleAxisY(tilt_range_y),
            SingleAxisX(tilt_range_x),
        ]
    )


def no_wedge() -> NoWedge:
    """Create a no-wedge missing wedge model."""
    return NoWedge()



from __future__ import annotations
from pathlib import Path
from typing import (
    Any,
    Iterable,
    TYPE_CHECKING,
    Literal,
    Sequence,
    Union,
    overload,
)
import numpy as np
from numpy.typing import ArrayLike, NDArray
import polars as pl
from scipy.spatial.transform import Rotation
from acryo._types import nm
from acryo.molecules._group import MoleculeGroup
from acryo.molecules._cut import MoleculeCutGroup
from acryo.molecules._rotation import (
    axes_to_rotator,
    from_euler_xyz_coords,
    translate_euler,
)

if TYPE_CHECKING:
    from typing_extensions import Self, TypeGuard
    from polars.type_aliases import ParquetCompression

_CSV_COLUMNS = ["z", "y", "x", "zvec", "yvec", "xvec"]
PathLike = Union[str, Path]
IntoExpr = Union[str, pl.Expr]


class Molecules:
    """
    Object that represents position- and orientation-defined molecules.

    Positions are represented by a (N, 3) ``np.ndarray`` and orientations
    are represented by a ``scipy.spatial.transform.Rotation`` object.
    Features of each molecule can also be recorded by the ``features``
    property.

    Parameters
    ----------
    pos : ArrayLike
        Moleculs positions.
    rot : scipy.spatial.transform.Rotation object
        Molecule orientations.
    features : dataframe, optional
        Molecule features.
    """

    def __init__(
        self,
        pos: ArrayLike,
        rot: Rotation | None = None,
        features: pl.DataFrame | ArrayLike | dict[str, ArrayLike] | None = None,
    ):
        _pos = np.atleast_2d(pos).astype(np.float32)

        if _pos.shape[1] != 3:
            raise ValueError("Shape of pos must be (N, 3).")

        if rot is None:
            nmol = _pos.shape[0]
            if nmol > 0:
                quat = np.stack([np.array([0, 0, 0, 1])] * _pos.shape[0], axis=0)
            else:
                quat = np.zeros((0, 4))
            rot = Rotation.from_quat(quat)
        elif not isinstance(rot, Rotation):
            raise TypeError(f"`rot` must be a Rotation object, got {type(rot)}.")
        elif _pos.shape[0] != len(rot):
            raise ValueError(
                f"Length mismatch. There are {_pos.shape[0]} molecules but {len(rot)} "
                "rotation were given."
            )

        self._pos = _pos
        self._rotator = rot
        self._features: pl.DataFrame | None = None
        self.features = features

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(n={len(self)})"

    @classmethod
    def empty(cls, feature_labels: Iterable[str] = ()):
        return cls(
            pos=np.zeros((0, 3)),
            rot=None,
            features=pl.DataFrame(None, schema=list(feature_labels)),
        )

    @classmethod
    def from_axes(
        cls,
        pos: np.ndarray,
        z: np.ndarray | None = None,
        y: np.ndarray | None = None,
        x: np.ndarray | None = None,
    ) -> Self:
        """Construct molecule cloud with orientation from two of their local axes."""
        pos = np.atleast_2d(pos)

        if sum((_ax is not None) for _ax in [z, y, x]) != 2:
            raise TypeError("You must specify two out of z, y, and x.")

        # NOTE: np.cross assumes vectors are in xyz order. However, all the arrays here
        # are defined in zyx order. To build right-handed coordinates, we must invert
        # signs when using np.cross.
        if z is None:
            if x is None or y is None:
                raise TypeError("Two of x, y, z is needed.")
            x = np.atleast_2d(x)
            y = np.atleast_2d(y)
            z = cross(x, y, axis=1)
        elif y is None:
            if x is None or z is None:
                raise TypeError("Two of x, y, z is needed.")
            z = np.atleast_2d(z)
            x = np.atleast_2d(x)
            y = cross(z, x, axis=1)

        rotator = axes_to_rotator(z, y)
        return cls(pos, rotator)

    @classmethod
    def from_euler(
        cls,
        pos: np.ndarray,
        angles: ArrayLike,
        seq: str = "ZXZ",
        degrees: bool = False,
        order: str = "xyz",
        features: pl.DataFrame | None = None,
    ) -> Self:
        """Create molecules from Euler angles."""
        if order == "xyz":
            rotator = from_euler_xyz_coords(angles, seq, degrees)
        elif order == "zyx":
            rotator = Rotation.from_euler(seq, angles, degrees)
        else:
            raise ValueError("'order' must be 'xyz' or 'zyx'.")
        return cls(pos, rotator, features)

    @classmethod
    def from_quat(
        cls,
        pos: np.ndarray,
        quat: ArrayLike,
        features: pl.DataFrame | None = None,
    ) -> Self:
        """Create molecules from quaternions."""
        return cls(pos, Rotation.from_quat(quat), features)

    @classmethod
    def from_rotvec(
        cls,
        pos: np.ndarray,
        vec: ArrayLike,
        features: pl.DataFrame | None = None,
    ) -> Self:
        """Create molecules from rotation vectors."""
        return cls(pos, Rotation.from_rotvec(vec), features)

    @classmethod
    def from_matrix(
        cls,
        pos: np.ndarray,
        matrix: np.ndarray,
        features: pl.DataFrame | None = None,
    ) -> Self:
        """Create molecules from rotation matrices."""
        return cls(pos, Rotation.from_matrix(matrix), features)

    @classmethod
    def from_file(
        cls,
        path: PathLike,
        pos_cols: list[str] = ["z", "y", "x"],
        rot_cols: list[str] = ["zvec", "yvec", "xvec"],
    ) -> Self:
        """Load a file as a Molecules object using proper reader."""
        path = Path(path)
        if path.suffix in (".pq", ".parquet"):
            return cls.from_parquet(path, pos_cols, rot_cols)
        return cls.from_csv(path, pos_cols, rot_cols)

    @classmethod
    def from_csv(
        cls,
        path: PathLike,
        pos_cols: list[str] = ["z", "y", "x"],
        rot_cols: list[str] = ["zvec", "yvec", "xvec"],
        **pl_kwargs,
    ) -> Self:
        """Load csv as a Molecules object."""
        df = pl.read_csv(path, **pl_kwargs)
        return cls.from_dataframe(df, pos_cols, rot_cols)

    @classmethod
    def from_parquet(
        cls,
        path: PathLike,
        pos_cols: list[str] = ["z", "y", "x"],
        rot_cols: list[str] = ["zvec", "yvec", "xvec"],
    ) -> Self:
        """Load parquet as a Molecules object."""
        df = pl.read_parquet(path)
        return cls.from_dataframe(df, pos_cols, rot_cols)

    @classmethod
    def from_dataframe(
        cls,
        df: pl.DataFrame,
        pos_cols: list[str] = ["z", "y", "x"],
        rot_cols: list[str] = ["zvec", "yvec", "xvec"],
    ) -> Self:
        """Construct Molecules object from a DataFrame."""
        pos_cols = pos_cols.copy()
        rot_cols = rot_cols.copy()
        pos = df.select(pos_cols)
        rotvec = df.select(rot_cols)
        cols = pos.columns + rotvec.columns
        feature_columns = [c for c in df.columns if c not in cols]
        if len(feature_columns) == 0:
            features = None
        else:
            features = df.select(feature_columns)
        return cls(
            pos.to_numpy(),
            Rotation.from_rotvec(rotvec.to_numpy()),
            features=features,
        )

    @classmethod
    def from_random(
        cls,
        pos: np.ndarray,
        seed: int | None = None,
        features: pl.DataFrame | None = None,
    ) -> Self:
        """Create randomly oriented molecules from given positions."""
        rot = Rotation.random(len(pos), random_state=seed)
        return cls(pos, rot, features=features)

    @property
    def features(self) -> pl.DataFrame:
        """Molecules features."""
        if self._features is None:
            import polars as pl

            return pl.DataFrame(None)
        return self._features

    @features.setter
    def features(self, value):
        if value is None:
            self._features = None
        else:
            if not isinstance(value, pl.DataFrame):
                df = pl.DataFrame(value)
            else:
                df = value
            if df.shape == (0, 0):
                self._features = None
                return None
            if len(df) != self.pos.shape[0]:
                raise ValueError(
                    f"Length mismatch. There are {self.pos.shape[0]} molecules but "
                    f"{len(df)} features were given."
                )
            self._features = df
        return None

    def to_dataframe(self) -> pl.DataFrame:
        """Convert coordinates, rotation and features into a single data frame."""
        if dup := set(self.features.columns).intersection(_CSV_COLUMNS):
            raise ValueError(
                "Duplication between feature columns and postion, rotation columns. "
                f"Features should not contain {dup!r}. Please rename them."
            )
        rotvec = self.rotvec().astype(np.float32)
        df = pl.DataFrame(
            {
                "z": self.pos[:, 0],
                "y": self.pos[:, 1],
                "x": self.pos[:, 2],
                "zvec": rotvec[:, 0],
                "yvec": rotvec[:, 1],
                "xvec": rotvec[:, 2],
            },
        )
        if self._features is not None:
            df = df.with_columns(list(self._features))
        return df

    def to_csv(self, save_path: PathLike, float_precision: int | None = 4) -> None:
        """
        Save molecules as a csv file.

        Parameters
        ----------
        save_path : PathLike
            Save path.
        float_precision : int, default is 4
            Float precision.
        """
        return self.to_dataframe().write_csv(
            str(save_path),
            float_precision=float_precision,
        )

    def to_parquet(
        self,
        save_path: PathLike,
        *,
        compression: ParquetCompression = "zstd",
        compression_level: int | None = 10,
    ) -> None:
        """
        Save molecules as a parquet file.

        Parameters
        ----------
        save_path : PathLike
            Save path.
        compression : str, default is "zstd"
            Compression method, by default "zstd".
        """
        return self.to_dataframe().write_parquet(
            str(save_path),
            compression=compression,
            compression_level=compression_level,
        )

    def to_file(self, save_path: PathLike) -> None:
        """Save molecules as a file."""
        save_path = Path(save_path)
        if save_path.suffix in (".pq", ".parquet"):
            return self.to_parquet(save_path)
        return self.to_csv(save_path)

    def count(self) -> int:
        """Return the number of molecules."""
        return self._pos.shape[0]

    def __len__(self) -> int:
        """Return the number of molecules."""
        return self.count()

    def __getitem__(self, key: int | slice | list[int] | NDArray[np.integer]) -> Self:
        return self.subset(key)

    @property
    def pos(self) -> NDArray[np.float32]:
        """Positions of molecules."""
        return self._pos

    @property
    def x(self) -> NDArray[np.float64]:
        """Vectors of x-axis."""
        return self._rotator.apply([0.0, 0.0, 1.0])

    @property
    def y(self) -> NDArray[np.float64]:
        """Vectors of y-axis."""
        return self._rotator.apply([0.0, 1.0, 0.0])

    @property
    def z(self) -> NDArray[np.float64]:
        """Vectors of z-axis."""
        return self._rotator.apply([1.0, 0.0, 0.0])

    @property
    def rotator(self) -> Rotation:
        """Return ``scipy.spatial.transform.Rotation`` object"""
        return self._rotator

    @classmethod
    def concat(
        cls,
        moles: Iterable[Molecules],
        concat_features: bool = True,
        nullable: bool = True,
    ) -> Self:
        """Concatenate Molecules objects."""
        pos: list[np.ndarray] = []
        quat: list[np.ndarray] = []
        features: list[pl.DataFrame] = []
        for mol in moles:
            pos.append(mol.pos)
            quat.append(mol.quaternion())
            features.append(mol.features)

        all_pos = np.concatenate(pos, axis=0)
        all_quat = np.concatenate(quat, axis=0)
        if concat_features:
            how = "diagonal" if nullable else "vertical"
            all_features = pl.concat(features, how=how)
        else:
            all_features = None

        return cls(all_pos, Rotation(all_quat), features=all_features)

    def copy(self) -> Self:
        """Make a copy of the object."""
        pos = self.pos.copy()
        quat = self._rotator.as_quat(canonical=False).copy()
        if self._features is None:
            return self.__class__(pos, Rotation(quat))
        return self.__class__(pos, Rotation(quat), self._features)

    def subset(
        self,
        spec: int | slice | list[int] | NDArray[np.bool_] | NDArray[np.integer],
    ) -> Self:
        """
        Create a subset of molecules by slicing.

        Any slicing supported in ``numpy.ndarray``, except for integer, can be
        used here. Molecule positions and angles are sliced at the same time.

        Parameters
        ----------
        spec : int, slice, list of int, or ndarray
            Specifier that defines which molecule will be used. Any objects
            that numpy slicing are defined are supported. For instance,
            ``[2, 3, 5]`` means the 2nd, 3rd and 5th molecules will be used
            (zero-indexed), and ``slice(10, 20)`` means the 10th to 19th
            molecules will be used.

        Returns
        -------
        Molecules
            Molecule subset.
        """
        if isinstance(spec, int):
            if spec < 0:
                raise IndexError("Negative indexing is not supported.")
            if spec >= len(self):
                raise IndexError("Index out of range.")
            _spec = slice(spec, spec + 1)
        else:
            _spec = spec
        pos = self.pos[_spec]
        quat = self._rotator.as_quat(canonical=False)[_spec]
        if self._features is None:
            return self.__class__(pos, Rotation(quat))
        if _is_boolean_array(_spec):
            # NOTE: polars does not support boolean indexing
            return self.__class__(pos, Rotation(quat), self._features.filter(_spec))
        return self.__class__(pos, Rotation(quat), self._features[_spec])

    def affine_matrix(
        self, src: np.ndarray, dst: np.ndarray | None = None, inverse: bool = False
    ) -> NDArray[np.float32]:
        """
        Construct affine matrices using positions and angles of molecules.

        Parameters
        ----------
        src : np.ndarray
            Source coordinates.
        dst : np.ndarray, optional
            Destination coordinates. By default the coordinates of molecules will be used.
        inverse : bool, default is False
            Return inverse mapping if true.

        Returns
        -------
        (N, 4, 4) array
            Array of concatenated affine matrices.
        """
        if dst is None:
            dst = self.pos

        nmole = len(self)

        if inverse:
            mat = self._rotator.inv().as_matrix()
        else:
            mat = self.matrix()
        rot_mat = np.zeros((nmole, 4, 4), dtype=np.float32)
        rot_mat[:, 3, 3] = 1.0
        rot_mat[:, :3, :3] = mat

        translation_0 = np.stack([np.eye(4, dtype=np.float32)] * nmole, axis=0)
        translation_1 = np.stack([np.eye(4, dtype=np.float32)] * nmole, axis=0)
        translation_0[:, :3, 3] = dst
        translation_1[:, :3, 3] = -src

        return np.einsum("nij,njk,nkl->nil", translation_0, rot_mat, translation_1)

    def local_coordinates(
        self,
        shape: tuple[int, int, int],
        scale: nm = 1.0,
        *,
        squeeze: bool = True,
    ) -> NDArray[np.float32]:
        """
        Generate local coordinates at the neighborhood of each molecule.

        Parameters
        ----------
        shape : (int, int, int)
            Shape of the local coordinates.
        scale : float, default is 1.0
            Scale of the local coordinates.
        squeeze : bool, default is True
            If ture and the number of molecules is 1, the first axis will be removed.

        Returns
        -------
        (N, D, Z, Y, X) array
            World coordinates. ``array[i]`` can be directly used as the coordinates
            for ``ndi.map_coordinates``.
        """
        coords_list: list[NDArray[np.float32]] = []
        all_vec_x = self.x.astype(np.float32)
        all_vec_y = self.y.astype(np.float32)
        for index in range(self.count()):
            center = [s / 2 - 0.5 for s in shape]
            vec_x = all_vec_x[index]
            vec_y = all_vec_y[index]
            vec_z = cross(vec_x, vec_y)
            ind_z, ind_y, ind_x = (
                (np.arange(s, dtype=np.float32) - c) for s, c in zip(shape, center)
            )  # type: ignore
            x_ax: NDArray[np.float32] = vec_x[:, np.newaxis] * ind_x
            y_ax: NDArray[np.float32] = vec_y[:, np.newaxis] * ind_y
            z_ax: NDArray[np.float32] = vec_z[:, np.newaxis] * ind_z

            coords = (
                z_ax[:, :, np.newaxis, np.newaxis]
                + y_ax[:, np.newaxis, :, np.newaxis]
                + x_ax[:, np.newaxis, np.newaxis, :]
            )
            shifts = self.pos[index] / scale
            coords += shifts[:, np.newaxis, np.newaxis, np.newaxis]  # unit: pixel

            coords_list.append(coords)

        if len(coords_list) == 1 and squeeze:
            return coords_list[0]
        else:
            return np.stack(coords_list, axis=0)

    def matrix(self) -> NDArray[np.float64]:
        """
        Calculate rotation matrices that align molecules in such orientations
        that ``vec`` belong to the object.

        Returns
        -------
        (N, 3, 3) ndarray
            Rotation matrices. Rotations represented by these matrices transform
            molecules to the same orientations, i.e., align all the molecules.
        """
        return self._rotator.as_matrix()

    def euler_angle(
        self, seq: str = "ZXZ", degrees: bool = False
    ) -> NDArray[np.float64]:
        """
        Calculate Euler angles that transforms a source vector to vectors that
        belong to the object.

        Parameters
        ----------
        seq : str, default is "ZXZ"
            Copy of ``scipy.spatial.transform.Rotation.as_euler``. 3 characters
            belonging to the set {"X", "Y", "Z"} for intrinsic rotations, or
            {"x", "y", "z"} for extrinsic rotations. Adjacent axes cannot be the
            same. Extrinsic and intrinsic rotations cannot be mixed in one function
            call.
        degrees: bool, default is False
            Copy of ``scipy.spatial.transform.Rotation.as_euler``. Returned
            angles are in degrees if this flag is True, else they are in radians.

        Returns
        -------
        (N, 3) ndarray
            Euler angles.
        """
        seq = translate_euler(seq)
        return self._rotator.as_euler(seq, degrees=degrees)[..., ::-1]

    def quaternion(self, canonical: bool = False) -> NDArray[np.float64]:
        """
        Calculate quaternions that transforms a source vector to vectors that
        belong to the object.

        Returns
        -------
        (N, 4) ndarray
            Quaternions.
        """
        return self._rotator.as_quat(canonical=canonical)

    def rotvec(self) -> NDArray[np.float64]:
        """
        Calculate rotation vectors that transforms a source vector to vectors
        that belong to the object.

        Returns
        -------
        (N, 3) ndarray
            Rotation vectors.
        """
        return self._rotator.as_rotvec()

    def translate(self, shifts: ArrayLike, copy: bool = True) -> Self:
        """
        Translate molecule positions by ``shifts``.

        Shifts are applied in world coordinates, not internal coordinates of
        every molecules. If molecules should be translated in their own
        coordinates, such as translating toward y-direction of each molecules
        by 1.0 nm, use ``translate_internal`` instead. Translation operation
        does not convert molecule orientations.

        Parameters
        ----------
        shifts : (3,) or (N, 3) array
            Spatial shift of molecules.
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing
            instance.

        Returns
        -------
        Molecules
            Instance with updated positional coordinates.
        """
        coords = self._pos + np.asarray(shifts, dtype=np.float32)
        if copy:
            features = self._features
            if features is not None:
                features = pl.DataFrame(list(features))  # copy
            out = self.__class__(coords, self._rotator, features=features)
        else:
            self._pos = coords
            out = self
        return out

    def translate_internal(self, shifts: ArrayLike, *, copy: bool = True) -> Self:
        """
        Translate molecule positions internally by ``shifts``.

        Shifts are applied in the coordinates of each molecule. If molecules
        should be translated in world coordinates use ``translate`` instead.

        Parameters
        ----------
        shifts : (3,) or (N, 3) array
            Spatial shift of molecules.
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing
            instance.

        Returns
        -------
        Molecules
            Instance with updated positional coordinates.
        """
        world_shifts = self._rotator.apply(shifts)
        return self.translate(world_shifts, copy=copy)

    def translate_random(
        self,
        max_distance: nm,
        *,
        seed: int | None = None,
        copy: bool = True,
    ) -> Self:
        """
        Apply random translation to each molecule.

        Translation range is restricted by a maximum distance and translation
        values are uniformly distributed in this region. Different translations
        will be applied to different molecules.

        Parameters
        ----------
        max_distance : nm
            Maximum distance of translation.
        seed : int, optional
            Random seed, by default None
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing
            instance.

        Returns
        -------
        Molecules
            Translated molecules.
        """
        nmole = len(self)
        np.random.seed(seed)
        r = np.random.random(nmole) * max_distance
        theta = np.random.random(nmole) * 2 * np.pi
        phi = np.random.random(nmole) * np.pi
        np.random.seed(None)
        cos_phi = np.cos(phi)
        shifts = np.stack(
            [
                r * np.sin(phi),
                r * cos_phi * np.sin(theta),
                r * cos_phi * np.cos(theta),
            ],
            axis=1,
        )
        return self.translate(shifts, copy=copy)

    def rotate_by_rotvec_internal(self, vector: ArrayLike, copy: bool = True) -> Self:
        """
        Rotate molecules using internal rotation vector.

        Vector components are calculated in the molecule-coordinate.

        Parameters
        ----------
        vector : ArrayLike
            Rotation vector(s).
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing instance.

        Returns
        -------
        Molecules
            Instance with updated angles.
        """
        vector = np.atleast_2d(vector)
        vec_x = self.x
        vec_y = self.y
        vec_z = cross(vec_x, vec_y, axis=1)
        world_rotvec = (
            vec_z * vector[:, 0][:, np.newaxis]
            + vec_y * vector[:, 1][:, np.newaxis]
            + vec_x * vector[:, 2][:, np.newaxis]
        )
        return self.rotate_by_rotvec(world_rotvec, copy=copy)

    def rotate_by_matrix(self, matrix: ArrayLike, copy: bool = True) -> Self:
        """
        Rotate molecules using rotation matrices, **with their position unchanged**.

        Parameters
        ----------
        matrix : ArrayLike
            Rotation matrices, whose length must be same as the number of
            molecules.
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing
            instance.

        Returns
        -------
        Molecules
            Instance with updated orientation.
        """
        rotator = Rotation.from_matrix(matrix)
        return self.rotate_by(rotator, copy)

    def rotate_by_quaternion(self, quat: ArrayLike, copy: bool = True) -> Self:
        """
        Rotate molecules using quaternions, **with their position unchanged**.

        Parameters
        ----------
        quat : ArrayLike
            Rotation quaternion.
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing
            instance.

        Returns
        -------
        Molecules
            Instance with updated orientation.
        """
        rotator = Rotation.from_quat(quat)
        return self.rotate_by(rotator, copy)

    def rotate_by_euler_angle(
        self,
        angles: ArrayLike,
        seq: str = "ZXZ",
        degrees: bool = False,
        order: Literal["xyz", "zyx"] = "xyz",
        copy: bool = True,
    ) -> Self:
        """
        Rotate molecules using Euler angles, **with their position unchanged**.

        Parameters
        ----------
        angles: array-like
            Euler angles of rotation.
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing
            instance.

        Returns
        -------
        Molecules
            Instance with updated orientation.
        """
        if order == "xyz":
            rotator = from_euler_xyz_coords(angles, seq, degrees)
        elif order == "zyx":
            rotator = Rotation.from_euler(seq, angles, degrees)
        else:
            raise ValueError("'order' must be 'xyz' or 'zyx'.")
        return self.rotate_by(rotator, copy)

    def rotate_by_rotvec(self, vector: ArrayLike, copy: bool = True) -> Self:
        """
        Rotate molecules using rotation vectors, **with their position unchanged**.

        Parameters
        ----------
        vector: array-like
            Rotation vectors.
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing
            instance.

        Returns
        -------
        Molecules
            Instance with updated orientation.
        """
        rotator = Rotation.from_rotvec(vector)
        return self.rotate_by(rotator, copy)

    def rotate_by(self, rotator: Rotation, copy: bool = True) -> Self:
        """
        Rotate molecule with a ``Rotation`` object.

        Note that ``Rotation`` object satisfies following equation.

        >>> rot1.apply(rot2.apply(v)) == (rot1*rot2).apply(v)

        Parameters
        ----------
        rotator : Rotation
            Molecules will be rotated by this object.
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing instance.

        Returns
        -------
        Molecules
            Instance with updated orientation.
        """
        rot = rotator * self._rotator
        if copy:
            features = self._features
            if features is not None:
                features = pl.DataFrame(list(features))
            out = self.__class__(self._pos, rot, features=features)
        else:
            self._rotator = rot
            out = self
        return out

    def rotate_random(self, copy: bool = True, seed: int | None = None) -> Self:
        """
        Rotate molecules randomly.

        Parameters
        ----------
        copy : bool, default is True
            If true, create a new instance, otherwise overwrite the existing instance.
        seed : int, default is None
            Random seed to generate randomized rotators.

        Returns
        -------
        Molecules
            Instance with updated orientation.
        """
        rotator = Rotation.random(len(self), random_state=seed)
        return self.rotate_by(rotator, copy=copy)

    def linear_transform(
        self,
        shift: ArrayLike,
        rotator: Rotation,
        inv: bool = False,
    ) -> Self:
        """Shift and rotate molecules around their own coordinate."""
        rotvec = rotator.as_rotvec()
        if inv:
            shift_corrected = rotator.apply(shift, inverse=True)
            return self.rotate_by_rotvec_internal(-rotvec).translate_internal(
                shift_corrected
            )
        else:
            shift_corrected = rotator.apply(shift)
            return self.translate_internal(shift_corrected).rotate_by_rotvec_internal(
                rotvec
            )

    def concat_with(
        self,
        other: Molecules,
        /,
        nullable: bool = True,
    ) -> Self:
        """
        Concatenate the molecules with the others.

        Parameters
        ----------
        other : Molecules
            Molecules to be concatenated.
        nullable : bool, default is True
            If true, missing features in either of the molecules will be filled with
            ``null``. Raise error otherwise.
        """
        pos = np.concatenate([self.pos, other.pos], axis=0)
        rot = np.concatenate(
            [
                self.rotator.as_quat(canonical=False),
                other.rotator.as_quat(canonical=False),
            ],
            axis=0,
        )
        if len(self.features) == 0:
            feat = other.features
        elif len(other.features) == 0:
            feat = self.features
        else:
            how = "diagonal" if nullable else "vertical"
            feat = pl.concat([self.features, other.features], how=how)
        return self.__class__(pos, Rotation.from_quat(rot), features=feat)

    @overload
    def groupby(self, by: IntoExpr) -> MoleculeGroup[str]:
        ...

    @overload
    def groupby(
        self, by: Sequence[IntoExpr] | tuple[IntoExpr, ...]
    ) -> MoleculeGroup[tuple[str, ...]]:
        ...

    @overload
    def groupby(
        self, by: IntoExpr, *more_by: IntoExpr
    ) -> MoleculeGroup[tuple[str, ...]]:
        ...

    def groupby(self, by, *more_by):
        """Group molecules into sub-groups."""
        df = self.to_dataframe()
        return MoleculeGroup(df.group_by(by, *more_by, maintain_order=True))

    group_by = groupby  # alias

    def cutby(self, by: str, bins: list[float]) -> MoleculeCutGroup:
        """Cut molecules into sub-groups by binning."""
        cat_name = f".category"
        while cat_name in self.features.columns:
            cat_name = f".{cat_name}"
        feature = self.features[by]
        cat = feature.cut(bins, as_series=True).alias(cat_name)  # type: ignore
        df = self.to_dataframe().with_columns(cat)
        return MoleculeCutGroup(
            df.group_by(cat_name, maintain_order=True), label=cat_name
        )

    def filter(
        self,
        predicate: pl.Expr | str | pl.Series | list[bool] | np.ndarray,
    ) -> Self:
        """Filter molecules by its features."""
        df = self.to_dataframe()
        df_filt = df.filter(predicate)
        return self.__class__.from_dataframe(df_filt)

    def head(self, n: int = 10) -> Self:
        """Return the first n molecules."""
        df = self.to_dataframe()
        return self.__class__.from_dataframe(df.head(n))

    def tail(self, n: int = 10) -> Self:
        """Return the last n molecules."""
        df = self.to_dataframe()
        return self.__class__.from_dataframe(df.tail(n))

    def sample(self, n: int = 10, seed: int | None = None) -> Self:
        """Return n randomly sampled molecules."""
        df = self.to_dataframe()
        return self.__class__.from_dataframe(df.sample(n, seed=seed))

    def sort(
        self,
        by: IntoExpr | Iterable[IntoExpr],
        *more_by: IntoExpr,
        descending: bool = False,
    ) -> Self:
        """
        Return a new instance with sorted molecules and features.

        Parameters
        ----------
        by : str or Expr or sequence of them
            Column name or expression to sort by.
        descending : bool, default is False
            If true, sort in descending order.
        """
        df = self.to_dataframe()
        return self.__class__.from_dataframe(
            df.sort(by, *more_by, descending=descending)
        )

    def with_features(
        self,
        exprs: IntoExpr | Iterable[IntoExpr],
        *more_exprs: IntoExpr,
        **named_exprs: IntoExpr,
    ) -> Self:
        """Return a new instance with updated features."""
        return self.__class__(
            self.pos,
            self.rotator,
            features=self.features.with_columns(exprs, *more_exprs, **named_exprs),
        )

    def drop_features(self, columns: str | Sequence[str], *more_columns: str) -> Self:
        """Return a new instance with updated features."""
        return self.__class__(
            self.pos,
            self.rotator,
            features=self.features.drop(columns, *more_columns),
        )

    def append(self, other: Molecules) -> Self:
        """
        Mutably append molecules.

        Parameters
        ----------
        other : Molecules
            Molecules to be appended. They must only have the subset of features,
            otherwise this method will raise ValueError.

        Returns
        -------
        Molecules
            Same instance with updated values.
        """
        if not isinstance(other, Molecules):
            raise TypeError(f"Expected Molecules, got {type(other)}")
        pos = np.concatenate([self.pos, other.pos], axis=0)
        rot = np.concatenate(
            [
                self.rotator.as_quat(canonical=False),
                other.rotator.as_quat(canonical=False),
            ],
            axis=0,
        )

        if self.count() == 0:
            feat = other.features
        else:
            feat = pl.concat([self.features, other.features], how="diagonal")
            if len(feat.columns) != len(self.features.columns):
                extra = set(other.features.columns) - set(self.features.columns)
                raise ValueError(
                    f"Cannot append molecules with extra columns: {extra}."
                )
        self._pos = pos
        self._rotator = Rotation.from_quat(rot)
        self._features = feat
        return self


def _is_boolean_array(a: Any) -> TypeGuard[NDArray[np.bool_]]:
    if isinstance(a, pl.Series):
        return a.dtype is pl.Boolean
    else:
        return getattr(a, "dtype", None) == "bool"


def cross(x: ArrayLike, y: ArrayLike, axis=None) -> np.ndarray:
    """Vector outer product in zyx coordinate."""
    return -np.cross(x, y, axis=axis)  # type: ignore



from __future__ import annotations

from typing import TypeVar
import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.spatial.transform import Rotation

_T = TypeVar("_T", bound=np.number)


def axes_to_rotator(z: ArrayLike | None, y: ArrayLike) -> Rotation:
    """Determine the Rotation object that rotates the z-axis to z and the y-axis to y."""
    y0 = _normalize(np.atleast_2d(y))
    rot_y = _get_align_rotator([[0, 1, 0]], y0)
    if z is None:
        z0 = _extract_orthogonal(y0, np.array([[1, 0, 0]]))
    else:
        z0 = _extract_orthogonal(y0, _normalize(np.atleast_2d(z)))
    z0_trans = rot_y.apply(z0, inverse=True)
    rot_z = _get_align_rotator([[1, 0, 0]], z0_trans)
    return rot_y * rot_z


def _get_align_rotator(src, dst) -> Rotation:
    """R.apply(src) == dst. Both length must be 1."""
    if np.all(np.abs(src + dst) < 1e-6):
        # cross product cannot be used for antiparallel vectors
        return Rotation.from_matrix(-np.eye(3))
    cross = np.cross(src, dst)
    sin = norm = np.sqrt(np.sum(cross**2, axis=1, keepdims=True))
    cos = np.sum(src * dst, axis=1, keepdims=True)
    theta = np.arctan2(sin, cos)

    norm[norm == 0] = np.inf
    return Rotation.from_rotvec(cross / norm * theta)


def from_euler_xyz_coords(
    angles: ArrayLike, seq: str = "ZXZ", degrees: bool = False
) -> Rotation:
    """Create a rotator using zyx-coordinate system, from Euler angles."""
    seq = translate_euler(seq)
    angles = np.asarray(angles)
    return Rotation.from_euler(seq, angles[..., ::-1], degrees)


def translate_euler(seq: str) -> str:
    table = str.maketrans({"x": "z", "z": "x", "X": "Z", "Z": "X"})
    return seq[::-1].translate(table)


def _normalize(a: NDArray[_T]) -> NDArray[_T]:
    """Normalize vectors to length 1. Input must be (N, 3)."""
    return a / np.sqrt(np.sum(a**2, axis=1))[:, np.newaxis]


def _extract_orthogonal(a: NDArray[_T], b: NDArray[_T]) -> NDArray[_T]:
    """Extract component of b orthogonal to a."""
    a_norm = _normalize(a)
    return b - np.sum(a_norm * b, axis=1)[:, np.newaxis] * a_norm



from __future__ import annotations
from abc import abstractmethod

import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation
from acryo.tilt import _base, _utils


class SingleAxis(_base.TiltSeriesModel):
    def __init__(self, tilt_range: tuple[float, float] = (-60.0, 60.0)):
        _min, _max = tilt_range
        if _min >= _max:
            raise ValueError(f"Tilt range {tilt_range!r} does not satisfy min < max.")
        if _min < -90.0 or _max > 90.0:
            raise ValueError(f"Tilt range {tilt_range!r} is not between -90 and 90.")
        self._tilt_range = tilt_range

    @property
    def tilt_range(self) -> tuple[float, float]:
        """Range of tilt angles in degrees."""
        return self._tilt_range

    @abstractmethod
    def _get_norms(self) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
        """Return the two normal vectors that define the missing wedge."""

    def create_mask(
        self,
        rotator: Rotation = Rotation.identity(),
        shape: tuple[int, int, int] = (49, 49, 49),
    ) -> NDArray[np.float32]:
        """
        Create a binary mask that covers tomographical missing wedge.

        Note that the mask is not shifted to the center of the Fourier domain.
        ``np.fft.fftn(img) * mask`` will be the correct way to apply the mask.

        Parameters
        ----------
        rotator : Rotation
            The rotation object that describes the direction of the mask.
        shape : tuple of int
            The shape of the mask.

        Returns
        -------
        np.ndarray
            Missing wedge mask.
        """
        normal0, normal1 = self._get_norms()
        shape_vector = np.array(shape, dtype=np.float32)
        rotator_inv = rotator.inv()
        normal0 = rotator_inv.apply(normal0 * shape_vector)
        normal1 = rotator_inv.apply(normal1 * shape_vector)
        vectors = _utils.get_indices(shape)
        dot0 = vectors.dot(normal0)
        dot1 = vectors.dot(normal1)
        missing = dot0 * dot1 <= 0
        return missing

    def _mask_from_norms(
        self,
        normal0: NDArray[np.float32],
        normal1: NDArray[np.float32],
        rotator: Rotation,
        shape: tuple[int, int, int],
    ) -> NDArray[np.float32]:
        shape_vector = np.array(shape, dtype=np.float32)
        rotator_inv = rotator.inv()
        normal0 = rotator_inv.apply(normal0 * shape_vector)
        normal1 = rotator_inv.apply(normal1 * shape_vector)
        vectors = _utils.get_indices(shape)
        dot0 = vectors.dot(normal0)
        dot1 = vectors.dot(normal1)
        missing = dot0 * dot1 <= 0
        return missing


class SingleAxisY(SingleAxis):
    def _get_norms(self) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
        return _utils.get_norms_y(self._tilt_range)

    def __repr__(self) -> str:
        tilt = f"{self.tilt_range[0]:.1f}, {self.tilt_range[1]:.1f}"
        return f"SingleAxis<y>({tilt})"


class SingleAxisX(SingleAxis):
    def _get_norms(self) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
        return _utils.get_norms_x(self._tilt_range)

    def __repr__(self) -> str:
        tilt = f"{self.tilt_range[0]:.1f}, {self.tilt_range[1]:.1f}"
        return f"SingleAxis<x>({tilt})"



from __future__ import annotations

import numpy as np

from acryo._typed_scipy import fftn, ifftn, affine_transform
from acryo.molecules import Molecules
from acryo._utils import compose_matrices
from acryo._rotation import normalize_rotations
from acryo._types import nm, degree


class TomogramGenerator:
    """
    A tester class for subtomogram averaging and alignment.

    Given a template image, this class can create tomogram by assembling
    rotated, noisy template images gridwise. Molecules objects can also
    be sampled with arbitrary positional errors.
    """

    def __init__(
        self,
        template: np.ndarray,
        grid_shape: tuple[int, int] = (10, 10),
        rotations=None,
        noise_sigma: float = 1,
        seed: int = 0,
    ) -> None:
        self._template = template
        self._grid_shape = grid_shape
        self._quaternions = normalize_rotations(rotations)
        self._noise_sigma = noise_sigma
        self._seed = seed

    @property
    def template(self):
        return self._template

    @property
    def grid_shape(self):
        return self._grid_shape

    @property
    def quaternions(self):
        return self._quaternions

    @property
    def noise_sigma(self):
        return self._noise_sigma

    def _get_matrices(self, rng: np.random.Generator):
        gy, gx = self.grid_shape
        quat_idx = rng.choice(self.quaternions.shape[0], size=(gy * gx))

        from scipy.spatial.transform import Rotation

        rotators = [Rotation.from_quat(self.quaternions[idx]) for idx in quat_idx]
        return compose_matrices(np.array(self.template.shape) / 2 - 0.5, rotators)

    def get_tomogram(
        self, pad_width: int = 0, tilt_range: tuple[degree, degree] = (-90, 90)
    ) -> np.ndarray:
        rng = np.random.default_rng(self._seed)
        template = self.template
        if pad_width > 0:
            template = np.pad(template, pad_width, dims="zyx")  # type: ignore

        gy, gx = self.grid_shape

        mols: list[list[np.ndarray]] = [
            [template.copy() for _ in range(gy)] for _ in range(gx)
        ]
        if self.quaternions.shape[0] > 0:
            matrices = self._get_matrices(rng)
            mtx_iterator = iter(matrices)
            for i, j in wrange(gy, gx):
                mtx = next(mtx_iterator)
                mols[i][j] = affine_transform(mols[i][j], mtx)

        for i, j in wrange(gy, gx):
            mols[i][j] += rng.normal(scale=self.noise_sigma, size=template.shape)

        if tilt_range != (-90, 90):
            mw = _missing_wedge_mask(template.shape, tilt_range=tilt_range)
            for i, j in wrange(gy, gx):
                ft = np.fft.fftshift(fftn(mols[i][j]))
                mols[i][j] = np.real(np.fft.ifftshift(ifftn(ft * mw)))

        tomogram: np.ndarray = np.block(mols)  # type: ignore
        return tomogram

    def sample_molecules(self, max_distance: nm = 3.0, scale: nm = 1.0):
        gy, gx = self.grid_shape
        shape_vec = np.array(self.template.shape)
        offset = (shape_vec - 1) / 2 * scale
        vy, vx = shape_vec[1:] * scale
        centers = []
        for i, j in wrange(gy, gx):
            centers.append(offset + np.array([0.0, vy * i, vx * j]))
        centers = np.stack(centers, axis=0)
        return Molecules(centers).translate_random(
            max_distance=max_distance, seed=self._seed
        )


def _missing_wedge_mask(
    shape: tuple[int, int, int], tilt_range: tuple[degree, degree]
) -> np.ndarray:
    """
    Create a missing-wedge binary mask image.

    Mask created by this function should be multiplied to Fourier transformed
    image.

    Parameters
    ----------
    shape : tuple of int
        Shape of the output array.
    tilt_range : tuple[float, float]
        Tomogram tilt range in degree.

    Returns
    -------
    np.ndarray
        A binary mask.
    """
    radmin, radmax = np.deg2rad(tilt_range)
    x0 = (shape[2] - 1) / 2
    z0 = (shape[0] - 1) / 2
    zz, yy, xx = np.indices(shape)
    d0 = zz - z0 - np.tan(radmin) * (xx - x0)
    d1 = zz - z0 - np.tan(radmax) * (xx - x0)
    missing = d0 * d1 < 0
    return missing


def wrange(l0: int, l1: int):
    for i in range(l0):
        for j in range(l1):
            yield i, j



from typing import Union, TYPE_CHECKING
from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from scipy.spatial.transform import Rotation

RangeLike = tuple[float, float]
Ranges = tuple[RangeLike, RangeLike, RangeLike]
RotationType = Union[Ranges, "Rotation"]

# type alias
nm: TypeAlias = float
degree: TypeAlias = float
pixel: TypeAlias = int
subpixel: TypeAlias = float



# pyright: reportPrivateImportUsage=false

from __future__ import annotations
from typing import Generic, TypeVar, TYPE_CHECKING
import inspect
import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from ._base import BaseAlignmentModel, TemplateType

_T = TypeVar("_T", bound="BaseAlignmentModel")


class ParametrizedModel(Generic[_T]):
    """Parametrized, alignment model factory function."""

    def __init__(self, cls: type[_T], **params):
        self._cls = cls
        self._params = params

    def create_model(
        self, template: TemplateType, mask: NDArray[np.float32] | None = None
    ) -> _T:
        bound = inspect.signature(self._cls).bind_partial(**self._params)
        return self._cls(template, mask, *bound.args, **bound.kwargs)

    def __call__(
        self, template: TemplateType, mask: NDArray[np.float32] | None = None
    ) -> _T:
        return self.create_model(template, mask)

    def __repr__(self) -> str:
        _args = ", ".join(f"{k}={v!r}" for k, v in self._params.items())
        return f"ParametrizedModel[{self._cls.__name__}]({_args})"

    @property
    def quaternions(self) -> NDArray[np.float32]:
        from acryo._rotation import normalize_rotations

        rotations = self._params.get("rotations", None)
        return normalize_rotations(rotations)

    @property
    def has_rotation(self) -> bool:
        """If the alignment model to be created has rotation optimization."""
        return self.quaternions.shape[0] > 1



# pyright: reportPrivateImportUsage=false

from __future__ import annotations
import os
from typing import Any, NamedTuple, Callable, TypeVar
import numpy as np
from dask import array as da


class ImageData(NamedTuple):
    image: da.Array
    scale: float


Loader = Callable[[str], tuple[np.memmap, float]]
_L = TypeVar("_L", bound=Loader)


class ImageReaderRegistry:
    def __init__(self):
        self._reader: dict[str, Loader] = {}

    def register(self, *ext: str) -> Callable[[_L], _L]:
        ext_list: list[str] = []
        for _ext in ext:
            if not _ext.startswith("."):
                _ext = "." + _ext
            ext_list.append(_ext)

        def _register(f: _L) -> _L:
            nonlocal ext_list
            for ext in ext_list:
                self._reader[ext] = f
            return f

        return _register

    def imread(self, path: str, chunks: Any = "auto") -> ImageData:
        _, ext = os.path.splitext(path)
        img, scale = self._reader[ext](path)
        return ImageData(image=as_dask(img, chunks=chunks), scale=scale)

    def imread_array(self, path: str) -> tuple[np.ndarray, float]:
        _, ext = os.path.splitext(path)
        img, scale = self._reader[ext](path)
        return np.asarray(img, dtype=np.float32), scale


REG = ImageReaderRegistry()


def imread(path: str, chunks: Any = "auto") -> ImageData:
    return REG.imread(path, chunks=chunks)


def register(*ext: str) -> Callable[[_L], _L]:
    return REG.register(*ext)


# TODO: check scale unit
@REG.register(".tif", ".tiff")
def open_tif(path: str):
    try:
        from tifffile import TiffFile
    except ImportError as e:
        ext = os.path.splitext(path)[1]
        e.msg = (
            f"No module named tifffile. To read {ext[1:]} files, please\n"
            "$ pip install tifffile"
        )
        raise e

    with TiffFile(path) as tif:
        pagetag = tif.series[0].pages[0].tags  # type: ignore

        tags = {v.name: v.value for v in pagetag.values()}
        scale = tags["XResolution"][1]

        img: np.memmap = tif.asarray(out="memmap")  # type: ignore
    return img, scale


@REG.register(".mrc", ".rec", ".map")
def open_mrc(path: str):
    try:
        import mrcfile
    except ImportError as e:
        ext = os.path.splitext(path)[1]
        e.msg = (
            f"No module named mrcfile. To read {ext[1:]} files, please\n"
            "$ pip install mrcfile"
        )
        raise e

    with mrcfile.mmap(path, mode="r") as mrc:
        scale: float = mrc.voxel_size.x / 10
        img: np.memmap = mrc.data  # type: ignore

    return img, scale


def as_dask(mmap: np.memmap, chunks: Any = "auto") -> da.Array:
    img = da.from_array(mmap, chunks=chunks, meta=np.array([])).map_blocks(
        np.asarray, dtype=mmap.dtype
    )  # type: ignore
    return img
